package edu.unika.aifb.owltools.evaluation;

import java.util.*;

import org.semanticweb.kaon2.api.*;

/**
 * Performs an ontology evaluation. This is an abstract superclass that
 * implements several common tasks for different evaluations.
 */
public abstract class Evaluation {
	
	/**
	 * Creates an evaluation of an ontology.
	 * @param ontologyToBeEvaluated The ontology to be evaluated
	 * @param URI The URI of the evaluation. If set to null, a random URI is generated
	 * @param contextOfEvaluation The context ontology for the evaluation. If set to null, no context is being used
	 */
	Evaluation(Ontology ontologyToBeEvaluated, String URI, Ontology contextOfEvaluation) { initialize(ontologyToBeEvaluated, URI, contextOfEvaluation); }
	
	/**
	 * Human-readable name of the Evaluation. Needs to be implemented by all evaluations
	 * 
	 * @return Name of the evaluation
	 */
	abstract public String evaluationClassName();
	
	/**
	 * URI for the class of evaluations. Needs to be implemented by all evaluations
	 * 
	 * @return URI of the evaluation
	 */
	abstract public String evaluationClassURI();
	
	/**
	 * Human-readable description of what this evaluation is doing. Needs to
	 * be implemented by all evaluations
	 * 
	 * @return Description of the evaluation type
	 */
	abstract public String evaluationClassDescription();
	
	
	/**
	 * Returns the URI of the evaluation
	 * @return URI of the evaluation
	 */
	public String getURI() { return uri; }
	
	/**
	 * An evaluation always return a value in [0,1]
	 * @return Evaluation measure
	 */
	public double getEvaluation() { return grade; }
	
	/**
	 * Returns all the identified problems of the Evaluation
	 * @return A map of problem URIs and the respective problems.
	 */
	public Map<String, Problem> getProblemMap() { return problems; }
	
	/**
	 * Returns all the identified problems of the Evaluation
	 * @return The set of problems
	 */
	public Collection<Problem> getProblems() { return problems.values(); }
	
	/**
	 * Human readable description of how the evaluation fared
	 * @return Description of the evaluation
	 */
	public String getDescription() { return description; }
	
	/**
	 * The ontology being evaluated
	 * @return Evaluated ontology
	 */
	public Ontology getOntology() { return ontology; }
	
	/**
	 * The context of the evaluation
	 * @return Evaluation context
	 */
	public Ontology getContext() { return context; }
	
	/**
	 * Has there been a succesful evaluation?
	 * @return true if the evaluation was succesful, else false
	 */
	public boolean succesfulEvaluation() { return success; }
	
	/**
	 * The base namesapce for the generated evaluation URIs
	 */
	public String evaluationURIBase = "http://owltools.ontoware.org/evaluation/";

	/**
	 * Implements the actual evaluation
	 * @return true if the evaluation was succesful, else false
	 */
	abstract protected boolean evaluate();

	/**
	 * Evaluated ontology
	 */
	protected Ontology ontology;
	
	/**
	 * Evaluation context
	 */
	protected Ontology context;
	
	/**
	 * Describes the evaluation result in a human readable form
	 */
	protected String description;
	
	/**
	 * The problems and their URIs
	 */
	protected Map<String, Problem> problems;
	
	/**
	 * Sets the evaluation grade. It is clipped into [0,1]. 0 means
	 * the ontology is really bad, 1 means it is perfect (with regards
	 * to this evaluation measure)
	 * @param evaluation intented evaluation grade
	 */
	protected void setEvaluation(double evaluation) {
		if (evaluation < 0.0)
			grade = 0.0;
		else if (evaluation > 1.0)
			grade = 1.0;
		else
			grade = evaluation;
	}

	/**
	 * Adds a problem to the set of problems about the ontology
	 * @param problem new problem
	 * @return the new problem map
	 */
	protected Map<String, Problem> addProblem(Problem problem) {
		problems.put(problem.getURI(), problem);
		return problems;
	}
	
	/**
	 * The evaluation's URI
	 */
	private String uri;
	
	/**
	 * Was the evaluation succesful?
	 */
	private boolean success;
	
	/**
	 * The evaluation result
	 */
	private double grade;

	/**
	 * A random variable, used for generating evaluation URIs
	 */
	static Random random = new Random();
	
	/**
	 * Creates a new random URI for the evaluation
	 * @return new evaluation URI
	 */
	private String getRandomURIEvaluation() {
		return getRandomURI("evaluation#e");
	}
	
	/**
	 * Creates a new random URI for a solution
	 * @return new evaluation URI
	 */
	protected String getRandomURISolution() {
		return getRandomURI("solution#s");
	}
	
	/**
	 * Creates a new random URI for a problem
	 * @return new evaluation URI
	 */
	protected String getRandomURIProblem() {
		return getRandomURI("problem#p");
	}
	
	/**
	 * Creates a new random URI for anything
	 * @return new evaluation URI
	 */
	private String getRandomURI(String mid) {
		return evaluationURIBase + mid + Math.abs(random.nextInt()); 
	}
	
	/**
	 * Initializes the Evaluation object, sets all the values and
	 * then also evaluates the ontology.
	 * @param URI Evaluation URI
	 * @param ontologyToBeEvaluated Ontology to be evaluated
	 * @param contextOfEvaluation Evaluation context
	 */
	private void initialize(Ontology ontologyToBeEvaluated, String URI, Ontology contextOfEvaluation) {
		if (null==URI)
			uri = getRandomURIEvaluation();
		else
			uri = URI;
		ontology = ontologyToBeEvaluated;
		context = contextOfEvaluation;
		setEvaluation(0.0);
		description = "No description";
		problems = new HashMap<String, Problem>();
		success = evaluate();
	}
}
