package edu.unika.aifb.owltools.evaluation;

import java.util.*;
import org.semanticweb.kaon2.api.*;

/**
 * A solution provided by an evaluation to a certain identified
 * problem.
 */
public class Solution {
	
	/**
	 * The URI of the solution
	 */
	private String uri;
	
	/**
	 * The ontology change events that make up the solution
	 */
	protected List<OntologyChangeEvent> solution;
	
	/**
	 * Human-readable description of the solution
	 */
	protected String description;
	
	/**
	 * How good is this solution deemed by the evaluation?
	 * The result is in [0,1]
	 */
	protected double preference;
	
	/**
	 * Creates a solution
	 * @param URI URI for the solution
	 * @param whatToChange A list of ontology change events that implement the solution
	 * @param whatItIs Describes the solution in a human readable way
	 * @param howGoodIsTheIdea Sets the preference for the solution. Clipped in [0,1]
	 */
	public Solution(String URI, List<OntologyChangeEvent> whatToChange, String whatItIs, double howGoodIsTheIdea) {
		uri = URI;
		solution = whatToChange;
		description = whatItIs;
		preference = howGoodIsTheIdea;
		if (preference>1.0) preference=1.0;
		if (preference<0.0) preference=0.0;
	}

	/**
	 * The URI for this solution
	 * @return URI for this solution
	 */
	public String getURI() { return uri; }
	
	/**
	 * The implementation of the solution
	 * @return A list of ontology change events that implement the solution
	 */
	public List<OntologyChangeEvent> getSolution() { return solution; }
	
	/**
	 * A human-readable description of the solution proposal
	 * @return Descripition of the solution proposal
	 */
	public String getDescription() { return description; }
	
	/**
	 * The preference for this solution. The preference is in [0,1],
	 * with 0 meaning that this solution probably won't do any good
	 * and 1 that it would lead to a perfect ontology.
	 * @return Preference for this solution
	 */
	public double getPreference() { return preference; }
	
	/**
	 * Applies this solution to an ontology
	 * @param ontology Problematic ontology
	 * @return Unproblematic ontology
	 * @throws KAON2Exception
	 */
	public Ontology apply(Ontology ontology) throws KAON2Exception {
		ontology.applyChanges(solution);
		return ontology;
	}
}
