package edu.unika.aifb.owltools.metrics;

import java.util.Random;
import org.semanticweb.kaon2.api.Ontology;

/**
 * Calculates a structural metric
 */
public abstract class Metric {
	
	/**
	 * Creates the measurement of an ontology.
	 * @param ontologyToBeMeasured The ontology to be measured
	 * @param URI The URI of the measurement. If set to null, a random URI is generated
	 * @param contextOfMeasurement The context ontology for the measurement. If set to null, no context is being used
	 */
	Metric(Ontology ontologyToBeMeasured, String URI, Ontology contextOfMeasurement) { initialize(ontologyToBeMeasured, URI, contextOfMeasurement); }
	/**
	 * Human-readable name of the Metric. Needs to be implemented by all metrics
	 * @return Name of the metric
	 */
	abstract public String metricClassName();
	
	/**
	 * URI for the metric class. Needs to be implemented by all metrics
	 * @return URI of the metric
	 */
	abstract public String metricClassURI();
	
	/**
	 * Human-readable description of what the metric is measuring. Needs to
	 * be implemented by all metrics
	 * @return Description of the metric
	 */
	abstract public String metricClassDescription();
	
	
	/**
	 * Returns the URI of the metric measurement
	 * @return URI of the metric measurement
	 */
	public String getURI() { return uri; }
	
	/**
	 * The result of the metric
	 * @return Metric measurement
	 */
	public double getResult() { return result; }
	
	/**
	 * The ontology being measured
	 * @return Measured ontology
	 */
	public Ontology getOntology() { return ontology; }
	
	/**
	 * The context of the measurement
	 * @return Measurement context
	 */
	public Ontology getContext() { return context; }
	
	/**
	 * Has there been a succesful measurement?
	 * @return true if the measurement was succesful, else false
	 */
	public boolean succesfulMeasurement() { return success; }
	
	/**
	 * The base namesapce for the generated measurement URIs
	 */
	public String metricsURIBase = "http://owltools.ontoware.org/metrics/";

	/**
	 * Implements the actual measurement
	 * @return true if the evaluation was succesful, else false
	 */
	abstract protected boolean measure();

	/**
	 * Measured ontology
	 */
	protected Ontology ontology;
	
	/**
	 * Measurement context
	 */
	protected Ontology context;
	
	/**
	 * The measurement's URI
	 */
	private String uri;
	
	/**
	 * Was the measurement succesful?
	 */
	private boolean success;
	
	/**
	 * The measurement result
	 */
	protected double result;
	
	/**
	 * Contains the error message in case the measuring fails
	 * @return Error message. If empty, no error message was saved.
	 */
	public String getError() { return error; }
	
	/**
	 * A possible error message when measuring failes
	 */
	protected String error = "";

	/**
	 * A random variable, used for generating URIs
	 */
	static Random random = new Random();
	
	/**
	 * Creates a new random URI for the measurement
	 * @return new evaluation URI
	 */
	private String getRandomURIMeasurement() {
		return getRandomURI("measurement#m");
	}
	
	/**
	 * Creates a new random URI for anything
	 * @return new evaluation URI
	 */
	private String getRandomURI(String mid) {
		return metricsURIBase + mid + Math.abs(random.nextInt()); 
	}
	
	/**
	 * Initializes the Metrics object, sets all the values and
	 * then also measures the ontology.
	 * @param URI Measurement URI
	 * @param ontologyToBeMeasured Ontology to be measured
	 * @param context Measurement context
	 */
	private void initialize(Ontology ontologyToBeMeasured, String URI, Ontology contextOfMeasurement) {
		if (null==URI)
			uri = getRandomURIMeasurement();
		else
			uri = URI;
		ontology = ontologyToBeMeasured;
		context = contextOfMeasurement;
		success = measure();
	}
	

}
