package edu.unika.aifb.owltools.serializer;

import java.util.*;

import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.owl.elements.*;

import edu.unika.aifb.owltools.tools.DLPConvert.IntegrityConstraintHandling;
import edu.unika.aifb.owltools.utils.Printer;

/**
 * Serializes a set of rules to an XSB Prolog program.
 */
public class PrologSerializer extends ProgramSerializer {
	// TODO: this list of predicate names that are predefined by XSB is extremely incomplete
	static final String m_reservednames[]={"assert", "attribute", "consult", "integer", "length", "name", "number", "phrase", "seeing", "throw"};
	
	// should all non-Prolog rules be dropped (commented out)?
	protected boolean m_forceprolog;
	// should explicit (F-Logic-style) quantifiers be added
	protected boolean m_fquantifiers;

	private boolean m_head = false;
	private boolean m_emptyBody = false;
	private int m_body = 0;
	private String m_rulestring;
	private String m_quantifierstring;
	private Set<String> m_foundvars;
	
	/**
	 * Constructor
	 * @param printer is the printer used for output
	 */
	public PrologSerializer(Printer printer) {
		super(printer);
		// Java's crappy way of implementing default parameters:
		m_forceprolog = true;
		m_fquantifiers = false;
	}
	
	/**
	 * Constructor
	 * @param printer is the printer used for output
	 * @param forceprolog if set to false, non-Prolog rules will be printed as well (especially disjunctive rules)
	 */
	public PrologSerializer(Printer printer, boolean forceprolog, boolean fquantifiers) {
		super(printer);
		m_forceprolog = forceprolog;
		m_fquantifiers = fquantifiers;
	}

	public Object visit(Rule rule) {
		m_rulestring = "";
		m_quantifierstring = "";
		m_foundvars = new HashSet<String>();
		//Printer.verbose("rule: " + rule.toString());
		int headLength = rule.getHeadLength();
		int bodyLength = rule.getBodyLength();
		m_emptyBody = (bodyLength == 0);

		if (headLength == 0)
			m_emptyHead = true;

		m_head = true;
		for (int i = 0; i < headLength; ++i) {
			m_rulestring += (i==0 ? "" : ", ");
			rule.getHeadLiteral(i).accept(this);
		}
		m_head = false;

		if (!m_emptyBody) {
			while (m_rulestring.length() < 30) { // simple aligned indenting
				m_rulestring = " " + m_rulestring;
			}
			m_rulestring += " :- ";
			for (int i = 0; i < bodyLength; ++i) {
				m_rulestring += (i==0 ? "" : ", ");
				m_body = 0;
				rule.getBodyLiteral(i).accept(this);
				if (m_stat) {
					m_rulestring += ", N is ";
					for (int k = 0; k < m_body; ++k) {
						if (k != 0)
							m_rulestring += " + ";
						m_rulestring += "N" + k;
					}
					if (headLength > 1)
						m_rulestring += " + 1";
				}
			}
		}

		if (headLength > 1) {
			if (m_forceprolog) {
				m_rulestring = "%%% Disjunctive rule %%% " + m_rulestring;
			}
			m_DDL = true;
		}

		if ( !m_quantifierstring.equals("") ) {
			m_quantifierstring += " ";
			while (m_quantifierstring.length() < 11) { // simple aligned indenting
				m_quantifierstring += " ";
			}
		}
		m_printer.putline(m_quantifierstring + m_rulestring + ".");
		return null;
	}

	public Object visit(Literal literal) {

		// negation
		if (!literal.isPositive()) {
			m_rulestring += "NOT ";
			m_negation = true;
		}

		// predicate
		String predname = processName(literal.getPredicate().toString());
		// escape illegal names
		for (int i=0; i<m_reservednames.length; i++) {
			if (m_reservednames[i].equals(predname)) {
				predname = predname + "__";
				continue;
			}
		}
		if (predname.startsWith("_")) {
			predname = "ot__" + predname; // escape
		}
		m_rulestring += predname;

		// parameters
		int arity = literal.getArity();
		if (arity > 0) m_rulestring += "(";
		for (int i = 0; i < arity; ++i) {
			m_rulestring += (i==0 ? "" : ", ");
			literal.getArgument(i).accept(this);
		}
		if (m_stat) {
			if (m_head) {
				if (m_emptyBody)
					m_rulestring += ", 0";
				else
					m_rulestring += ", N";
			} else {
				m_rulestring += ", N" + (m_body++);
			}
		}
		if (arity > 0) m_rulestring += ")";
		return null;
	}

	public Object visit(Constant constant) {
		m_rulestring += processDataConstant(constant.toString());
		return null;
	}

	public Object visit(Individual individual) {
		m_rulestring += processName(individual.toString());
		return null;
	}

	public Object visit(Variable variable) {
		m_rulestring += variable.toString();
		if ( m_fquantifiers && !m_foundvars.contains(variable.toString()) ) {
			if ( m_quantifierstring.equals("") ) {
				m_quantifierstring = "FORALL " + variable.toString();
			} else {
				m_quantifierstring += ", " + variable.toString();
			}
			m_foundvars.add(variable.toString());
		}
		return null;
	}

	public Object visit(FunctionalTerm fsym) {
		m_rulestring += processName(fsym.getFunctionSymbol().toString());
		if (fsym.getArity() > 0) {
			m_rulestring += "(";
			for (int i=0; i<fsym.getArity(); i++) {
				fsym.getArgument(i).accept(this);
			}
			m_rulestring += ")";
		}
		return null;
	}

	/**
	 * Change constants to become Prolog compatible: remove ^^types as
	 * appropriate, and maybe escape problematic symbols.
	 */
	private String processDataConstant(String constant) {
		String result;
		// check for datatype addon
		int i = constant.lastIndexOf("^^");
		if (i != -1) {
			result = constant.substring(0,i); // strip type information
			String type = constant.substring(i+2);
			// TODO: are all XSD float and double notations accepted by XSB Prolog?
			if ( (type.equals("<xsd:int>")) || (type.equals("<xsd:long>")) || (type.equals("<xsd:short>")) || (type.equals("<xsd:byte>")) || (type.equals("<xsd:integer>")) || (type.equals("<xsd:unsignedInt>")) || (type.equals("<xsd:unsignedLong>")) || (type.equals("<xsd:unsignedShort>")) || (type.equals("<xsd:unsignedByte>")) || (type.equals("<xsd:positiveInteger>")) || (type.equals("<xsd:nonPositiveInteger>")) || (type.equals("<xsd:negativeInteger>")) || (type.equals("<xsd:nonNegativeInteger>")) || (type.equals("<xsd:decimal>")) || (type.equals("<xsd:float>")) || (type.equals("<xsd:double>")) ) { // remove quotes "..."
				if ( (result.charAt(0) == '"') && (result.charAt(result.length()-1) == '"') ) {
					result = result.substring(1,result.length()-1);
				}
			}
		} else {
			result = constant;
		}
		return result;
	}

	public void start(String filename) {
		super.start(filename);
		if (m_forceprolog && !m_fquantifiers) {
			m_printer.putline("% :- auto_table. %% automatic tabling in XSB to ensure termination (as far as possible)");
			m_printer.putline("% :- suppl_table. %% automatic tabling in XSB to improve performance");
		}
	}

}
