package edu.unika.aifb.owltools;

import edu.unika.aifb.owltools.tools.*;
import org.semanticweb.kaon2.api.*;

import java.util.*;

/**
 * All functionality of the owl tools is bundled here for simple usage in own programs
 */
public class api {
	/**
	 * Holds the dump object that is used for some of the functionality 
	 */
	private static Dump m_dump = null;
	
	/**
	 * Counts the number of axioms in an ontology
	 * @param ontology The ontology to be counted
	 * @return The number of axioms in the ontology
	 * @throws KAON2Exception
	 */
	static public int count(Ontology ontology) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.count(ontology);
	}

	/**
	 * Counts the number of axioms of a specific kind in an ontology
	 * @param ontology The ontology to be counted
	 * @param filter The kind of axioms to be counted. Must be one of org.semanticweb.kaon2.api.owl.axioms
	 * @return The number of axioms of the kind
	 * @throws KAON2Exception
	 */
	static public int count(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.count(ontology, filter);
	}

	/**
	 * Counts the number of entities in an ontology
	 * @param ontology The ontology to be counted
	 * @return The number of entities in the ontology
	 * @throws KAON2Exception
	 */
	static public int countEntities(Ontology ontology) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.countEntities(ontology);
	}
	
	/**
	 * Counts the number of entities of a specific kind in an ontology
	 * @param ontology The ontology to be counted
	 * @param filter The kind of entities to be counted. Must be one of org.semanticweb.kaon2.api.owl.elements
	 * @return The number of entities of that kind
	 * @throws KAON2Exception
	 */
	static public int countEntities(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.countEntities(ontology, filter);		
	}

	/**
	 * Returns a list of all axioms in the ontology. The syntax is KAON2's own axiom syntax
	 * @param ontology The ontology to be listed
	 * @return A list of strings, each one representing one axiom
	 * @throws KAON2Exception
	 */
	static public List<String> dump(Ontology ontology) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.dump(ontology);
	}
	
	/**
	 * Returns a list of all axioms of a certain kind. The syntax is KAON2's own axiom syntax
	 * @param ontology The ontology to be listed from
	 * @param filter The kind of axioms to be listed. Must be one of org.semanticweb.kaon2.api.owl.axioms
	 * @return A list of strings, each one representing one axiom of the specified kind
	 * @throws KAON2Exception
	 */
	static public List<String> dump(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.dump(ontology, filter);
	}
	
	/**
	 * Returns a list of all entities mentioned in the ontology
	 * @param ontology The ontology to be listed from
	 * @return A list of strings, with the URIs of all entities in the ontology
	 * @throws KAON2Exception
	 */
	static public List<String> dumpEntities(Ontology ontology) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.dumpEntities(ontology);
	}
	
	 /**
	  * List all entities of a specified type
	  * @param ontology The ontology to be listed from
      * @param filter The kind of entities to be counted. Must be one of org.semanticweb.kaon2.api.owl.elements
	  * @return A list of strings, each one an URI of an entity of the specified kind
	  * @throws KAON2Exception
	  */
	static public List<String> dumpEntities(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_dump) m_dump = new Dump();
		return m_dump.dumpEntities(ontology, filter);
	}

	/**
	 * Holds the filter object that is used for the filtering 
	 */
	private static Filter m_filter = null;
	
	
	/**
	 * Filters an ontology by only leaving axioms of a certain kind inside the ontology
	 * @param ontology The ontology to be filtered
	 * @param filter The axiomtype to remain in the ontology
	 * @return An ontology with only the axioms of the specified kind
	 * @throws KAON2Exception
	 */
	static public Ontology filter(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_filter) m_filter = new Filter();
		return m_filter.filter(ontology, filter);
	}
	
	/**
	 * Filters an ontology be removing all axioms of a certain kind from the ontology
	 * @param ontology The ontology to be filtered
	 * @param filter The axiomtype to be removed from the ontology
	 * @return An ontology with all axioms of the specified kind removed
	 * @throws KAON2Exception
	 */
	static public Ontology filterOut(Ontology ontology, Class filter) throws KAON2Exception {
		if (null == m_filter) m_filter = new Filter();
		return m_filter.filterOut(ontology, filter);
	}

	/**
	 * Holds the diff object that is used to create an ontology diff 
	 */
	private static Diff m_diff = null;

	/**
	 * Gets the difference between two ontologies. It returns all the axioms that are part
	 * of the diff ontology but not the base ontology.
	 * @param base The ontology to compare to
	 * @param diff The ontology with the extra axioms
	 * @param result The ontology that holds all the extra axioms in diff but not in base
	 * @throws KAON2Exception
	 */
	static public void diff(Ontology base, Ontology diff, Ontology result) throws KAON2Exception {
		if (null == m_diff) m_diff = new Diff();
		m_diff.diff(base, diff, result);
	}

	/**
	 * Holds the latex object that is used to create the latex serialization 
	 */
	private static Latex m_latex = null;
	
	/**
	 * Creates a latex serialization of an ontology. By using the latex class itself you
	 * have much more possibilities to customize the output 
	 * @param ontology The ontology to be tex'ed
	 * @return The result, being a list of strings that represent lines of a latex program
	 * @throws KAON2Exception
	 */
	static public List<String> latex(Ontology ontology) throws KAON2Exception {
		if (null == m_latex) m_latex = new Latex();
		return m_latex.latex(ontology, true);
	}

	/**
	 * Holds the luna object that is used to add the local una axiom 
	 */
	private static Luna m_luna = null;

	/**
	 * Adds an axiom to an ontology that makes all the individuals in the ontology
	 * mutually different.
	 * @param ontology Ontology to be locally una'd.
	 * @return The ontology with the local una axiom
	 * @throws KAON2Exception
	 */
	static public Ontology luna(Ontology ontology) throws KAON2Exception {
		if (null == m_luna) m_luna = new Luna();
		return m_luna.luna(ontology);
	}
	
	/**
	 * Holds the merge object that is used to merge two ontologies 
	 */
	private static Merge m_merge = null;

	/**
	 * Merges two ontologies. It adds up the axioms from both input ontologies
	 * @param one One of the ontologies 
	 * @param two The other ontology
	 * @param result The merged ontology
	 * @throws KAON2Exception
	 */
	static public void merge(Ontology one, Ontology two, Ontology result) throws KAON2Exception {
		if (null == m_merge) m_merge = new Merge();
		m_merge.merge(one, two, result);
	}

	/**
	 * Holds the meta object that is used to create meta ontologies and recompile them
	 * to actual ones 
	 */
	private static Meta m_meta = null;
	
	/**
	 * Creates a metaontology of an ontology based on the OWL Metamodel. By using the
	 * Meta class directly much more options are available.
	 * @param ontology The ontology to be meta'd
	 * @param metaontology The returned metaontology of the given ontology
	 * @throws KAON2Exception
	 */
	static public void meta(Ontology ontology, Ontology metaontology) throws KAON2Exception {
		if (null == m_meta) m_meta = new Meta();
		m_meta.meta(ontology, metaontology, false);
	}
	
	/**
	 * Creates a concrete ontology from a metaontology
	 * @param metaontology The metaontology that holds the concrete ontology
	 * @param ontology The returned concrete ontology
	 * @throws KAON2Exception
	 */
	static public void demeta(Ontology metaontology, Ontology ontology) throws KAON2Exception, InterruptedException {
		if (null == m_meta) m_meta = new Meta();
		m_meta.demeta(metaontology, ontology);
	}

	/**
	 * Holds the populate object that is used to populate ontologies
	 */
	private static Populate m_populate = null;
	
	/**
	 * Populates an ontology with random instances
	 * @param ontology The ontology to be populated
	 * @param instances The number of instances to be added
	 * @return The populated ontology
	 * @throws KAON2Exception
	 */
	static public Ontology populate(Ontology ontology, int instances) throws KAON2Exception {
		if (null == m_populate) m_populate = new Populate();
		return m_populate.populate(ontology, instances);
	}

	/**
	 * Holds the satisfiable object that is used to check ontologies for satisfiability
	 */
	private static Satisfiable m_satisfiable = null;

	/**
	 * Checks if an ontology is logically satisfiable
	 * @param ontology The ontology to be checked
	 * @return true if the ontology is satisfiable, else false
	 * @throws KAON2Exception
	 * @throws InterruptedException
	 */
	static public boolean satisfiable(Ontology ontology) throws KAON2Exception, InterruptedException {
		if (null == m_satisfiable) m_satisfiable = new Satisfiable();
		return m_satisfiable.satisfiable(ontology);
	}
}
