package edu.unika.aifb.owltools.evaluation;

import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.elements.*;

/**
 * Checks if in an ontology all entities have a label.
 * This class can also be used to implement own evaluations. It is documented
 * heavily, so that it can be reused as a template for further evaluation measures.
 */
public class AllLabeled extends Evaluation {
	/**
	 * Constructor. Keep it the same, just change the constructor name appropriately
	 * @param ontologyToBeEvaluated Input ontology. This is the ontology to be evaluated. It is accessible as ontology
	 * @param URI URI for the evaluation. This is just the name for this specific evaluation. You don't have to know it, or change it.
	 * @param contextOfEvaluation Evaluation context. Also all further parameters that you need for the evaluation should be handled over the context. Accessible as context. Null if none is given
	 */
	public AllLabeled(Ontology ontologyToBeEvaluated, String URI, Ontology contextOfEvaluation) { super(ontologyToBeEvaluated, URI, contextOfEvaluation); }
	
	/**
	 * Human-readable name for the evaluation type. Give it a telling name that could
	 * be used in an user interface.
	 */
	public String evaluationClassName() { return "Labeled entities"; };
	
	/**
	 * URI for the evaluation type. Don't mix it with the URI for the
	 * evaluation itself. This is the kind of evaluation. Just give it some
	 * URI in your own namespace.
	 */
	public String evaluationClassURI() { return "http://owltools.ontoware.org/evaluation/type#alllabeled"; }
	
	/**
	 * Human-readable description of the evaluation type. Give it a short, but useful
	 * description, that could be displayed to users in an appropriate interface
	 * to help users decide whether to use the evaluation or not.
	 */
	public String evaluationClassDescription() { return "Checks if all the entitites in the ontology have a label."; }
	
	/**
	 * Checks if in an ontology all entities have a label. This is the meat of your
	 * evaluation function. You have to set all results here, and do all the computing.
	 */
	protected boolean evaluate() {
		try {
			AnnotationProperty labelProperty = KAON2Manager.factory().annotationProperty("http://www.w3.org/2000/01/rdf-schema#label");
			Request<Entity> request = ontology.createEntityRequest();
			Set<Entity> entities = request.get();
			int labeled = 0;
			int unlabeled = 0;
			Set<Entity> nolabels = new HashSet<Entity>();
			for (Entity entity : entities) {
				if (!OWLClass.class.isInstance(entity)) continue;
				OWLEntity owlEntity = (OWLEntity)entity;
				Object label = owlEntity.getAnnotationValue(ontology, labelProperty);
				if (label==null) {
					unlabeled++;
					nolabels.add(entity);
				} else {
					labeled++;
				}
			}
			// The formula assures that the evaluation result is in [0,1].
			// Remember that 0 means "really bad" and 1 means "perfect".
			// setEvaluation is the only way to set the evaluation result, and
			// it assures that the evaluation is in the given range.
			if ((labeled+unlabeled)==0)
				setEvaluation(1.0);
			else
				setEvaluation((labeled*1.0) / ((labeled+unlabeled)*1.0));

			if (unlabeled > 0) {
				// there have been unlabeled entitites. The only solution to this
				// problem is to label them. So let us create a solution.
				// We need first a description of the solution
				String howToSolve = "Give labels to the following entities: ";
				for (Entity nolabel : nolabels)
					howToSolve += nolabel.getURI() + ", ";
				howToSolve = howToSolve.substring(0, howToSolve.length()-2);
				// Now let us create the solution, that we have a description of
				// how to solve. Pitily, we cannot automatically solve the problem.
				Solution solution = new Solution(
						getRandomURISolution(), // the solution needs an URI
						null, // there is no automatic way to resolve th problem
						howToSolve, // the description we created above is entered here
						0.9); // applying this solution leads to a perfect score -- but it is cumbersome
				
				howToSolve = "Use the owl label tool"; // Who says this is advertising?
				Solution solution2 = new Solution(
						getRandomURISolution(),
						null,
						howToSolve,
						1.0);
				
				Map<String,Solution> solutions = new HashMap<String,Solution>();
				solutions.put(solution.getURI(), solution);			
				solutions.put(solution2.getURI(), solution2);			
				
				// Now we can describe the problem
				Problem problem = new Problem(
						getRandomURIProblem(), // the problem needs a name
						null, // the problem is not localizable in certain axioms in the ontology
						"There are unlabeled entities", // this is the description for the problem
						solutions, // the solution we created above
						1.0-getEvaluation()); // the badness of this problem is reverse to the quality of the ontology
				addProblem(problem);
			}
		} catch (KAON2Exception k) {
			// Evaluation does not throw any exceptions. So make sure to catch them
			// all here and provide the feedback on why it failed in the description
			// string. Return false to show that the evaluation has failed.
			description = k.toString();
			return false;
		}
		// The evaluation has went well, everything is fine, return true
		return true;
	}

}
