package edu.unika.aifb.owltools.evaluation;

import java.util.*;

import org.semanticweb.kaon2.api.*;

/**
 * A problem that is identified by an ontology evaluation
 */
public class Problem {

	/**
	 * The URI of the problem
	 */
	private String uri;
	
	/**
	 * The part of the ontology that includes the problem.
	 */
	protected Ontology problematicPart;
	
	/**
	 * Human-readable description of the problem.
	 */
	protected String description;
	
	/**
	 * A number of possible solutions. Mapping from their URIs to
	 * the solutions.
	 */
	protected Map<String, Solution> solutions;
	
	/**
	 * How bad is the problem (in [0, 1]
	 */
	protected double severity;
	
	/**
	 * Creates a Problem
	 * @param URI the URI of the problem.
	 * @param whichPartIsWrong The part of the ontology that has the problem.
	 * @param whatIsWrong Description of the problem.
	 * @param possibleSolutions Possible solutions to the problem.
	 * @param howBadIsIt The severity of the problem, clipped to [0,1].
	 */
	public Problem(String URI, Ontology whichPartIsWrong, String whatIsWrong, Map<String, Solution> possibleSolutions, double howBadIsIt) {
		uri = URI;
		problematicPart = whichPartIsWrong;
		description = whatIsWrong;
		solutions = possibleSolutions;
		severity = howBadIsIt;
		if (severity<0.0) severity = 0.0;
		if (severity>1.0) severity = 1.0;
	}
	
	/**
	 * The URI of the problem
	 * @return URI
	 */
	public String getURI() { return uri; }
	
	/**
	 * Which part of the ontology is problematic? Can be used to
	 * highlight the part.
	 * @return Problemaitc part
	 */
	public Ontology getProblematicPart() { return problematicPart; }
	
	/**
	 * Human readable description of the problem
	 * @return Human readable description of the problem
	 */
	public String getDescription() { return description; }
	
	/**
	 * How bad is the problem? In [0, 1], where 0 means no problem
	 * at all, and 1 means a real showstopper. The severity of the
	 * problem is not comparable between evaluation types.
	 * @return Severity of the problem
	 */
	public double getSeverity() { return severity; }
	
	/**
	 * Returns possible solutions to the problem. May be empty
	 * @return A map of solutions with their URIs
	 */
	public Map<String, Solution> getSolutionMap() { return solutions; }
	
	/**
	 * Returns possible solutions to the problem. May be empty
	 * @return All possible solutions
	 */
	public Collection<Solution> getSolutions() { return solutions.values(); }
	
	/**
	 * Returns the solution that is rated best by the evaluation. If
	 * no solution is available, null is returned.
	 * @return Best rated solution. May be null.
	 */
	public Solution getBestSolution() {
		Solution solution = null;
		double bestPreference = -1.0;
		for (Solution s : solutions.values())
			if (s.getPreference() > bestPreference) {
				solution = s;
				bestPreference = s.getPreference();
			}
		return solution;
	}
	
	/**
	 * Applies the best solution to the provided ontology.
	 * Should be the ontology that was evaluated. If no solution is
	 * available, no changes are performed
	 * @param ontology The problematic ontology
	 * @return The changed ontology
	 * @throws KAON2Exception
	 */
	Ontology applyBestSolution(Ontology ontology) throws KAON2Exception {
		if (solutions.isEmpty()) return ontology;
		return getBestSolution().apply(ontology);
	}
	
	/**
	 * Applies the given solution to the provided ontology.
	 * Should be the ontology that was evaluated. If no solution with
	 * the URI is available, no changes are performed.
	 * @param URI Chosen solution 
	 * @param ontology Problematic ontology
	 * @return Ontology with applied solution
	 * @throws KAON2Exception
	 */
	Ontology applySolution(String URI, Ontology ontology) throws KAON2Exception {
		if (solutions.containsKey(URI))
			return solutions.get(URI).apply(ontology);
		else
			return ontology;
	}
}
