package edu.unika.aifb.owltools.shell;

import java.awt.event.*;
import java.util.*;

/**
 * this is a helper class for the shell. it circumvents
 * java's impossibility of catching each keystroke and thus
 * offer autocompletion possibilities
 */
public class ConsoleKeyListener extends KeyAdapter {
	
	/**
	 * the window with the shell
	 */
	ShellWindow window;
	
	/**
	 * all commands entered by the user
	 */
	List<String> history;
	
	/**
	 * position when browsing the history
	 */
	int historypos = 0;
	
	/**
	 * words for autocompletion
	 */
	SortedSet<String> words;
	
	/**
	 * is the tab used right now?
	 */
	int tabbed = -1;
	
	/**
	 * what did the user really type?
	 */
	String lastFragment = "";
	
	/**
	 * what was the last key event?
	 */
	int lastKeyEvent;
	
	/**
	 * Constructor for the listener
	 * @param calling the Shell Window that registers this listeners
	 */
	public ConsoleKeyListener(ShellWindow calling) {
		super();
		window = calling;
		history = new LinkedList<String>();
	}
	
	/**
	 * catches the key event
	 */
	public void keyPressed (KeyEvent keyEvent) {
		switch (keyEvent.getKeyCode()) {
			case KeyEvent.VK_UP :
				keyEvent.consume();
				if (historypos>0) {
					historypos--;
					window.setline(history.get(historypos));
				}
				break;
			case KeyEvent.VK_DOWN :
				keyEvent.consume();
				if (historypos<(history.size()-1)) {
					historypos++;
					window.setline(history.get(historypos));
				}
				break;
			case KeyEvent.VK_ESCAPE :
				keyEvent.consume();
				window.clearline(); 
				break;
			case KeyEvent.VK_ENTER :
				keyEvent.consume();
				eval();
				break;
			case KeyEvent.VK_LEFT :
				if (window.cursorX() < (window.prompt.length()+1))
					keyEvent.consume();
				break;
			case KeyEvent.VK_BACK_SPACE :
				if (window.cursorX() < (window.prompt.length()+1))
					keyEvent.consume();
				break;
			case KeyEvent.VK_TAB :
				keyEvent.consume();
				autocomplete();
				break;
			default:
				// TextArea will take care of displaying a regular character
		}
		lastKeyEvent = keyEvent.getKeyCode();
	}
	
	/**
	 * sends the typed command to the shell evaluator
	 *
	 */
	void eval() {
		String cmd = window.lastline();
		cmd = cmd.substring(4, cmd.length());
		if (history.size()==0)
			history.add(cmd);
		if (!cmd.equals(history.get(history.size()-1)))
			history.add(cmd);
		historypos = history.size();
		window.eval(cmd);
	}
	
	/**
	 * implements the autocompletion
	 *
	 */
	void autocomplete() {
		// here you could get the context to become a bit more intelligent
		if (lastKeyEvent == KeyEvent.VK_TAB) {
			// tab pressed several times on same fragment
			tabbed++;
		} else {
			// the first tab
			tabbed = 0;
			String cmd = window.lastline();
			cmd = cmd.substring(4, cmd.length());
			String[] words = cmd.split("(\\s)+");
			String lasttoken = words[words.length-1];
			if (cmd.endsWith(" ")) lasttoken = "";
			lastFragment = lasttoken;
		}
		String completion = getWord(lastFragment, tabbed);
		window.replaceLastWord(completion);
	}
	
	/**
	 * gets possible completions
	 * @param fragment what is really typed
	 * @param hit which one of the possible completions? (is being modulo'd)
	 * @return
	 */
	String getWord(String fragment, int hit) {
		if (words == null) {
			words = new TreeSet<String>();
			words.addAll(window.shell.commands.keySet()); 
			words.addAll(window.shell.axioms.keySet()); 
		}
		words.addAll(window.shell.ontologies.values());
		words.addAll((window.shell.ontologies.keySet()));
		//for (Ontology ontology : )
		// TODO fill words
		SortedSet<String> tail = words.tailSet(fragment);
		if (tail.isEmpty())
			return fragment;
		String completion = tail.first();
		if (!completion.startsWith(fragment))
			return fragment;
		if (hit==0)
			return completion;
		List<String> fits = new LinkedList<String>();
		for (String s : tail)
			if (s.startsWith(fragment))
				fits.add(s);
			else
				break;
		return fits.get(hit % fits.size());
	}
}
