package edu.unika.aifb.owltools.tools;

import java.util.*;
import java.io.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.axioms.ObjectPropertyAttribute;
import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.formatting.*;

import edu.unika.aifb.owltools.serializer.*;
import edu.unika.aifb.owltools.serializer.ProgramSerializer.*;
import edu.unika.aifb.owltools.utils.*;
import edu.unika.aifb.owltools.api;

/**
 * DLPconvert is a tool to transform different syntactical representations of
 * a DLP program into each other (refer to Raphael Volz, 2004). This is the main
 * class that calls the actual transformation algorithms provided in KAON2 and in
 * the other classes of this package.
 */
public class DLPConvert extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "dlpconvert"; }

	/**
	 * Specifies the different possible formats that dlpconvert is able to serialise
	 * set of rules in.
	 */
	public enum Format {
		/**
		 * The format directly usable by prolog implementations. This is set as the standard
		 * format.
		 */
		Prolog, 
		/**
		 * The RuleML format
		 */
		RuleML,
		/**
		 * The KAON2 Rules format
		 */
		Rule,
		/**
		 * FLogic, can be directly used by OntoBroker and other FLogic-engines
		 */
		FLogic,
		/**
		 * proposed SWRL extension to OWL
		 */
		SWRL
	};
	
	/**
	 * Possibilities of how to handle integrity constraints. Integrity constraints are
	 * rules with an empty head, that never must evaluate to true
	 */
	public enum IntegrityConstraintHandling {
		/**
		 * Integrity constraints are rendered as rules with an empty head. Some
		 * implementations cannot deal with such a result
		 */
		asIC,
		/**
		 * Integrity constraints are simply skipped. The result is weaker than the input
		 */
		noIC,
		/**
		 * A new predicate is included. After the evaluation of the results, it can be
		 * checked if the predicate is true. If so, an integrity constraint was violated.
		 */
		ICpredicate
	}
	
	/**
	 * The format for the output.
	 */
	private Format m_format = Format.Prolog;
	
	/**
	 * Should the ABox be transformed as well?
	 */
	private boolean m_aBox = true;
	
	/**
	 * Should the rule base be unfolded? This is a KAON2 internal thing
	 */ 
	private boolean m_folding = true;
	
	/**
	 * Do we allow for disjunctive heads in the rules? Note that some serializations
	 * actually treat several heads as conjunctive heads. The output here is meant
	 * to be disjunctive and thus, if this is set on, some outputs are not conformant
	 * to standards.
	 */
	private boolean m_allowDDL = false;
	
	/**
	 * Should disjunctive heads be transformed to conjunctie ones? This leads to
	 * changed semantics of the program and usually better performance. Refer to
	 * Hitzler, Vrandecic at the ISWC2005
	 */
	private boolean m_screech = false;
	
	/**
	 * if switched on, some statistics during the screech process are counted
	 */
	private boolean m_stat = false;
	
	/**
	 * How should names be serialized? Most implementations cannot deal with URIs, so
	 * a few different serizalization styles are included.
	 */
	private NameSerializationStyle m_nameStyle = NameSerializationStyle.uri;
	
	/**
	 * The URI of the ontology to be converted
	 */
	private String m_fileuri = "";

	/**
	 * The filename for the output to be saved
	 */
	private String m_output = "";

	/**
	 * How to handel integrity constraints
	 */
	private IntegrityConstraintHandling m_ichandling = IntegrityConstraintHandling.asIC; 
	
	/**
	 * this is used for the Tomcat based implementation, to count the usage of the tool
	 */
	private static int count = 0;

	/**
	 * Allows to start dlpconvert from the command line as well.
	 * Only used for a command line version.
	 * 
	 * @param args command line parameters
	 */
	public static void main(String[] args) {
		DLPConvert dlp = new DLPConvert();
		if (!dlp.processArguments(args)) return;
		dlp.compile();
		dlp.printer.close();
	}
	
	/**
	 * Processes the command line parameters and sets the appropriate flags instead.
	 * Only used for the command line version.
	 * 
	 * @param args command line parameters
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("dlpconvert")) || (s.equalsIgnoreCase("dlp")) || (s.equalsIgnoreCase("screech"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasFormat = false;
		boolean hasNameStyle = false;
		boolean hasABox = false;
		boolean hasFolding = false;
		boolean hasDDL = false;
		boolean hasVerbose = false;
		boolean hasIC = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				
				// Format
				if (arg.equalsIgnoreCase("-Prolog") || arg.equalsIgnoreCase("-datalog")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.Prolog;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-KAON2")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.Rule;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-RuleML")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.RuleML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-FLogic")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.FLogic;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SWRL")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.SWRL;
					hasFormat = true;
					unknownArg = false;
				}
				
				// Name serialization style
				if (arg.equalsIgnoreCase("-URI")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.uri;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-atom")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.atom;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-short")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.fragment;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ns") || (arg.equalsIgnoreCase("-Namespace")) ) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.ns;
					hasNameStyle = true;
					unknownArg = false;
				}
				
				// to ABox or not to ABox
				if (arg.equalsIgnoreCase("-abox")) {
					if (hasABox) { printUsageError(); return false; }
					m_aBox = true;
					hasABox = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noabox")) {
					if (hasABox) { printUsageError(); return false; }
					m_aBox = false;
					hasABox = true;
					unknownArg = false;
				}
				
				// folding
				if (arg.equalsIgnoreCase("-fold") || arg.equalsIgnoreCase("-unfold")) {
					if (hasFolding) { printUsageError(); return false; }
					m_folding = true;
					hasFolding = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-dontfold") || arg.equalsIgnoreCase("-nounfold")) {
					if (hasFolding) { printUsageError(); return false; }
					m_folding = false;
					hasFolding = true;
					unknownArg = false;
				}
				
				// DDL
				if (arg.equalsIgnoreCase("-ddl") || arg.equalsIgnoreCase("-full") || arg.equalsIgnoreCase("-x") || arg.equalsIgnoreCase("-xdlp")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = true;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noddl") || arg.equalsIgnoreCase("-dlp") || arg.equalsIgnoreCase("-noxdlp")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = false;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-screech")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = true;
					m_screech = true;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-screechcount") || arg.equalsIgnoreCase("-sc")) {
					if (hasDDL) { printUsageError(); return false; }
					m_stat = true;
					m_allowDDL = true;
					m_screech = true;
					hasDDL = true;
					unknownArg = false;
				}
				
				// Integrity Constraint handling
				if (arg.equalsIgnoreCase("-ic")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.asIC;
					hasIC = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noic")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.noIC;
					hasIC = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-icpredicate") || arg.equalsIgnoreCase("-icp")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.ICpredicate;
					hasIC = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-verb")||arg.equalsIgnoreCase("-verbose")) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}

			} else {
				if (arg.equalsIgnoreCase("dlpconvert") || arg.equalsIgnoreCase("dlp")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Creates an ontology consisting of the rules that KAON2's conversion to disjunctive
	 * datalog creates out of the given ontology.
	 * 
	 * @param in The ontology to be created.
	 * @param out The ontology that will hold the resulting rulebase.
	 * @throws KAON2Exception
	 * @throws InterruptedException
	 */
	public void dlpconvert(Ontology in, Ontology out) throws KAON2Exception, InterruptedException {
		Clear.clear(out);
		Reasoner reasoner = in.createReasoner();
		Collection<Rule> rules = reasoner.getReductionToDisjunctiveDatalog(true, true, true, true);
		rules = addTransitivity(rules, in);
		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		for (Rule rule : rules)
			changes.add(new OntologyChangeEvent(rule, OntologyChangeEvent.ChangeType.ADD));
		out.applyChanges(changes);
	}
	
	/**
	 * Convenience interface. If you have a name for your result ontology, you can use this.
	 * 
	 * @param ontology Ontology to be converted
	 * @param uri the URI of the ontology to be created
	 * @param physicalUri where is the ontology physically saved
	 * @return the converted ontology
	 * @throws KAON2Exception from KAON2's reduction
	 * @throws InterruptedException from KAON2's reduction
	 */
	public Ontology dlpconvert(Ontology ontology, String uri, String physicalUri) throws KAON2Exception, InterruptedException {
        DefaultOntologyResolver resolver = new DefaultOntologyResolver();
        resolver.registerReplacement(uri, physicalUri);
        ontology.getConnection().setOntologyResolver(resolver);
		Ontology rv = ontology.getConnection().createOntology(uri, new HashMap<String,Object>());
		dlpconvert(ontology, rv);
		return rv;
	}

	/**
	 * Convenience interface, so you can easily work with ontologies without having to define
	 * a name for it. Useful for testing or for temporary results.
	 * 
	 * @param ontology Ontology to be converted
	 * @return Converted ontology
	 * @throws KAON2Exception from KAON2's reduction
	 * @throws InterruptedException from KAON2's reduction
	 */
	public Ontology dlpconvert(Ontology ontology) throws KAON2Exception, InterruptedException {
		return dlpconvert(ontology, "http://owltools.ontoware.org/dlpconvert/", "file:C:/temp/dlpconvert.xml");
	}

	// public static void dlpconvert(Ontology ontology) <- replaces the very ontology
	// public static void screech(Ontology in, Ontology out)
	// public static Ontology screech(Ontology ontology, String uri, String physicalUri)
	// public static void screech(Ontology ontology) <- replaces inline
	// public static void removeICConstraints(Ontology ontology) removes IC Constraints inline
	// breakDisjunctions(O
	
	/**
	 * Creates an ontology consisting of the rules that KAON2's converstion to
	 * disjunctive datalog creates. If there is a disjunction in the head,
	 * it is changed into a conjunction. The result may be applied for approximate
	 * reasoning.
	 * 
	 * @param ontology the ontology to be converted
	 * @return the screech program, as an ontology
	 * @throws KAON2Exception from KAON2
	 * @throws InterruptedException from KAON2
	 */
	public Ontology screech(Ontology ontology) throws KAON2Exception, InterruptedException {
		Reasoner reasoner = ontology.createReasoner();
		Collection<Rule> rules = reasoner.getReductionToDisjunctiveDatalog(true, true, true, true);
        DefaultOntologyResolver resolver = new DefaultOntologyResolver();
        resolver.registerReplacement("http://screech/","file:screechtest.xml");
        ontology.getConnection().setOntologyResolver(resolver);
		Ontology rv = ontology.getConnection().createOntology("http://screech/", new HashMap<String,Object>());
		int disjrulecounter = 0;
		int disjcounter = 0; 
		int iccounter = 0;
		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		for (Rule rule : rules) {
			if (rule.getHeadLength() == 0) {
				iccounter++;
			} else if (rule.getHeadLength() == 1) {
				//List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
				changes.add(new OntologyChangeEvent(rule, OntologyChangeEvent.ChangeType.ADD));
				//if (!rv.containsAxiom(rule, false))
					//rv.applyChanges(changes);					
			} else {
				disjrulecounter++;
				int l = rule.getHeadLength();
				disjcounter += l;
				List<Literal> body = new ArrayList<Literal>();
				int b = rule.getBodyLength();
				for (int i = 0; i < b; i++) {
					body.add(rule.getBodyLiteral(i));
				}
				for (int i = 0; i < l; i++) {
					//List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
					List<Literal> head = new ArrayList<Literal>();
					head.add(rule.getHeadLiteral(i));
					Rule approxrule = KAON2Manager.factory().rule(head, false, body);
					changes.add(new OntologyChangeEvent(approxrule, OntologyChangeEvent.ChangeType.ADD));
					//if (!rv.containsAxiom(approxrule, false))
						//rv.applyChanges(changes);
				}
			}
		}
		if (m_stat) printer.verbose(disjcounter + " disjunctions");
		if (m_stat) printer.verbose(", " + disjrulecounter + " disjunctive rules");
		if (m_stat) printer.verbose(", " + iccounter + " ICs");
		rv.setDuplicateAxiomsThrowException(false);
		rv.applyChanges(changes);
		if (m_stat) printer.verbose(", " + api.count(rv) + " Rules");
		return rv;
	}
	
	public Collection<Rule> addTransitivity(Collection<Rule> rules, Ontology ontology) throws KAON2Exception {
  		Cursor<ObjectPropertyAttribute> cursor = ontology.createAxiomRequest(ObjectPropertyAttribute.class).openCursor();
  		while (cursor.hasNext()) {
  			ObjectPropertyAttribute opa = cursor.next();
  			if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_TRANSITIVE) {
  				// build new tranisitivy rule and add it
  				Predicate predicate = opa.getObjectProperty();
  				Term x = KAON2Manager.factory().variable("X");
  				Term y = KAON2Manager.factory().variable("Y");
  				Term z = KAON2Manager.factory().variable("Z");
  				Term[] xz = new Term[2];
  				Term[] xy = new Term[2];
  				Term[] yz = new Term[2];
  				xz[0] = xy[0] = x;
  				xy[1] = yz[0] = y;
  				xz[1] = yz[1] = z;  				
  				Literal head = KAON2Manager.factory().literal(true, predicate, xz);
  				List<Literal> body = new LinkedList<Literal>();
  				body.add(KAON2Manager.factory().literal(true, predicate, xy));
  				body.add(KAON2Manager.factory().literal(true, predicate, yz));
  				Rule rule = KAON2Manager.factory().rule(head, body);
  				rules.add(rule);
  			}
  		}
		return rules;
	}
	
	public void compile() {
		try {
	        if (m_format==Format.SWRL) {
		        if (!m_output.equals("")) printer.setFile(m_output);
        		Ontology ontology = loadOntology(m_fileuri);
        		Ontology result = dlpconvert(ontology);
        		result.saveOntology(OntologyFileFormat.OWL_RDF, printer.writer, "UTF-8");
        		return;
	        }
		} catch (KAON2Exception e) {
			printer.error("Could not open or reduce ontology " + m_fileuri);
			printer.error(e.toString());
		} catch (InterruptedException i) {
			printer.error("Could not reduce ontology " + m_fileuri);
			printer.error(i.toString());
		} catch (IOException i) {
			printer.error("Could not save to ontology.");
			printer.error(i.toString());
		}

		try {	        
	        ProgramSerializer serializer;
	        
	        switch (m_format) {
        		case Rule     : serializer = new RuleSerializer(printer); break;
        		case RuleML   : serializer = new RuleMLSerializer(printer); break;
        		case FLogic   : serializer = new FLogicSerializer(printer); break;
        		default       : serializer = new PrologSerializer(printer);
	        }

	        serializer.setStyle(m_nameStyle);
	        serializer.start(m_output);

	        try {

	        	Ontology ontology = loadOntology(m_fileuri);
		
		        Reasoner reasoner = ontology.createReasoner();

		        Collection<Rule> rules;
		        if (m_allowDDL) {
		        	//System.out.println(m_allowDDL);
		        	//System.out.println(m_aBox);
		        	//System.out.println(m_folding);		        	
		        	//System.out.println(m_folding && m_aBox);		        	
		        	rules = reasoner.getReductionToDisjunctiveDatalog(m_aBox, m_folding, m_folding && m_aBox, true);
		        } else
		        	rules = reasoner.getReductionToDLP(m_aBox, m_folding, m_folding && m_aBox, true);
		        
		        rules = addTransitivity(rules, ontology);
		        		
	        	serializer.setScreech(m_screech);
	        	serializer.setStat(m_stat);
	        	serializer.setIC(m_ichandling);

	        	for (Rule rule : rules)
		        	serializer.visit(rule);
			        
		        if ((!m_allowDDL)&&(serializer.encounteredDDL()))
		        	throw new KAON2Exception("Encountered rule with disjunctive head.");
		        //if (serializer.emptyHead())
		        	//throw new KAON2Exception("Encountered rule with empty head.");
		        if (serializer.hasNegation())
		        	throw new KAON2Exception("Encountered negative predicate.");

			} catch (KAON2Exception e) {
				printer.error("Could not open or reduce ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not reduce ontology " + m_fileuri);
				printer.error(i.toString());
			}
			finally {
		        serializer.end();
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
		
	}
	
	//////////////////////////////////////////////////////////////////////////////
	//////////        
	//////////  Web Interface Methods        
	//////////        
	//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Creates a return string for the Web Interface of DLPConvert.
	 * Only used by the web interface.
	 * 
	 * @param printer used printer
	 * @return stringt to be put into the webpage
	 */
	static String createReturnString(Printer printer) {
		String output = printer.getOutput();
		if (printer.error.equals(""))
			return output;
		String errorheader = "<h2>Error</h2>";
		if (output.equals(""))
			return errorheader + printer.error;
		errorheader += "</pre>The following error occured. dlpconvert tried to continue its work nevertheless and below the errormessage you see the result, but it may be incorrect or incomplete.<br /><pre>";
		return errorheader + printer.error + "<hr />" + output; 
	}
	
	/**
	 * Serializes the result in Flogic.
	 * Only used by the Web Interface. 
	 * 
	 * @param uri input uri of the ontology to be converted
	 * @return the returnstring for the answer HTML-page
	 */
	public static String compileToFlogic(String uri) {
		Printer printer = new Printer();
		printer.setString();
		DLPConvert dlp = new DLPConvert();
		dlp.printer = printer;
		dlp.m_nameStyle = NameSerializationStyle.fragment;
		dlp.m_format = Format.FLogic;
		dlp.m_fileuri = uri;
		dlp.compile();
		return createReturnString(printer);
	}
	
	/**
	 * Creates a DLPConversion based on an URI.
	 * Only used by the Web Interface.
	 * 
	 * @param uri uri of the ontology to be converted
	 * @param dir where to archive the requeste
	 * @return the return string for the answering HTML-page
	 */
	public static String compileUri(String uri, String dir) {
		Printer printer = new Printer();
		printer.html = true;
		printer.setString();
		count += 1;
		try {
			String filename = dir + "/uri" + String.valueOf(count) + ".uri";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(uri);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		DLPConvert dlp = new DLPConvert();
		dlp.printer = printer;
		dlp.m_nameStyle = NameSerializationStyle.fragment;
		dlp.m_fileuri = uri;
		dlp.compile();
		return createReturnString(printer);
	}
	
	/**
	 * Reads a string as an ontology and then converts this ontology.
	 * Only used by the Web Interface
	 * 
	 * @param string the ontology in string form
	 * @param dir the archive directory
	 * @return the resulting string for the answering HTML-page
	 */
	public static String compileString(String string, String dir) {
        Printer printer = new Printer();
		printer.html = true;
		printer.setString();
        count += 1;
		try {
			String filename = dir + "/owl" + String.valueOf(count) + ".xml";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(string);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		try {
			// construct ontology from string
			KAON2Connection connection = KAON2Manager.newConnection();
			String uri = "http://dlp.aifb.uni-karlsruhe.de/dlpconvert/internal";
	        DefaultOntologyResolver resolver=new DefaultOntologyResolver();
        	resolver.registerReplacement(uri, "file:temp.owl");
	        connection.setOntologyResolver(resolver);
			Ontology ontology = connection.createOntology(uri, new HashMap<String,Object>());
			
			
		    // import String into ontology
			Reader text = new StringReader(string);
			//OWLImporter importer = new OWLImporter();
			OpenProgressListener listener = new DefaultOpenProgressListener();
			ontology.importContentsFrom(text, uri, listener);
		    
			// compile it the usual way
	        ProgramSerializer serializer = new PrologSerializer(printer);
	        serializer.setStyle(ProgramSerializer.NameSerializationStyle.fragment);
	        serializer.start("");
			//DLPConvert dlp = new DLPConvert();
	        org.semanticweb.kaon2.api.reasoner.Reasoner reasoner = ontology.createReasoner();
	        Collection<Rule> rules = reasoner.getReductionToDLP(true, true, true, true);
	        for (Rule rule : rules)
	        	serializer.visit(rule);
	        serializer.end();
			
			// return the compilated
			connection.close();
		} catch (KAON2Exception e) {
			printer.error(e.toString());
		} catch (InterruptedException i) {
			printer.error(i.toString());
		} catch (IOException io) {
			printer.error(io.toString());
		}
		// TODO refactor this whole file
		return createReturnString(printer);
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl dlpconvert <input_url> [<output_filename>] [-atom|-URI|-short] ...");
		// TODO erroresolving, just species
		System.out.println("");
		System.out.println("dlpconvert is a KAON2 based tool that transforms DLP statements from");
		System.out.println("one syntactical representation into another (may be shortened as 'dlp')");
		//System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read. Has to be in OWL/XML format.");
		//System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the transformation will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		//System.out.println("");
		System.out.println("Possible options (take one option from each group):");
		//System.out.println("");
		System.out.println(" -version  prints the version of the tool");
		System.out.println("");
		System.out.println(" -datalog  the output is as a datalog program (default)");
		System.out.println(" -FLogic   the output is FLogic");
		System.out.println(" -SWRL     the output is OWL+SWRL");
		System.out.println(" -RuleML   the output is RuleML");
		System.out.println(" -KAON2    the output is in the internal KAON2 format");
		System.out.println("");
		System.out.println(" -URI      the IDs are full qualified URIs (default)");
		System.out.println(" -atom     the names are transformed to legal XSB atoms (not with KAON2)");
		System.out.println(" -short    just the fragment identifiers  (most readable, not with KAON2)");
		// TODO System.out.println(" -ns       namespace and fragment identifier combined");
		System.out.println("");
//		System.out.println(" -ABox     processes the ABox (default)");
//		System.out.println(" -noABox   skips the ABox Processing");
//		System.out.println("");
//		System.out.println(" -unfold   unfolds the new function terms introduced in reduction (default)");
//		System.out.println(" -nounfold skips the unfolding");
//		System.out.println("");
		System.out.println(" -X        allows disjunctions in head (uses disjunctive datalog algorihtm)");
		System.out.println(" -DLP      disallows disjunctions in head (default)");
		System.out.println(" -screech  breaks disjunctions in head to multiple rules (sets DDL)");
		System.out.println("");
		System.out.println(" -IC        leaves integrity constraints in resulting program (default)");
		System.out.println(" -noIC      removes integrity constraints from result");
		System.out.println(" -ICPredicate (or -ICP) sets predicate 'Inconsistent' instead of empty head");
		System.out.println("");
		System.out.println(" -verbose  offers further information (off by default)");
	}
}
