package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;
import org.semanticweb.kaon2.api.owl.axioms.*;

/**
 * ded weakens ontologies by removing datatype range declarations. They can
 * cause trouble to KAON2.
 */
public class Ded extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "ded"; }
	
	/**
	 * The URI of the ontology to be weakened
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to save the ontology
	 */
	private OntologyFileFormat m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, the result will be supressed 
	 */	
	private boolean m_result = true;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Ded d = new Ded();
		if (!d.processArguments(args)) return;
		d.compile();
		d.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("ded"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.ABSTRACT_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("ded")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the ded owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
	        	ontology = ded(ontology);
	        	
	        	if (m_result)
	        		ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	
	public Ontology ded(Ontology ontology) throws KAON2Exception {
		
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
  		
  		Cursor<DataPropertyRange> rangeCursor = ontology.createAxiomRequest(DataPropertyRange.class).openCursor();
  		while (rangeCursor.hasNext()) {
  			DataPropertyRange a = rangeCursor.next();
	        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
	        printer.verbose("Deleting: " + a.toString());
  		}
  		rangeCursor.close();
  		
  		//Cursor<DataPropertyDomain> domainCursor = ontology.createAxiomRequest(DataPropertyDomain.class).openCursor();
  		//while (domainCursor.hasNext()) {
  		//	DataPropertyDomain a = domainCursor.next();
	    //    changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
	    //    verbose("Deleting: " + a.toString());
  		//}
  		//domainCursor.close();
  		
  		//Cursor<DataPropertyMember> memberCursor = ontology.createAxiomRequest(DataPropertyMember.class).openCursor();
  		//while (memberCursor.hasNext()) {
  		//	DataPropertyMember a = memberCursor.next();
	    //    changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
	    //    verbose("Deleting: " + a.toString());
  		//}
  		//memberCursor.close();

  		ontology.applyChanges(changes);

	  	return ontology;
	}
	
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl ded <input_url> [<output_filename>] [-verbose] [-rdf|-xml] [-no]");
		System.out.println("");
		System.out.println("ded weakens an ontology by removing all DataPropertyRange axioms.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the weakened ontology will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result. In combination with -verbose you");
		System.out.println("     get a list of the needed changes, without apllying them.");
	}	
}
