package edu.unika.aifb.owltools.tools;

import java.util.*;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.Ontology;
import edu.unika.aifb.owltools.evaluation.*;

/**
 * Accesses the evaluation package for the command line tools
 */
public class Evaluate extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "evaluate"; }
	
	/**
	 * The URI of the ontology to be weakened
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";

	/**
	 * Show solutions
	 */
	private boolean m_solutions = true; 
	
	/**
	 * Show problems
	 */
	private boolean m_problems = true;
	
	/**
	 * list of evaluations to be applied
	 */
	private List<String> m_evaluations = new LinkedList<String>(); 

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Evaluate eval = new Evaluate();
		if (!eval.processArguments(args)) return;
		eval.compile();
		eval.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("eval")) || (s.equalsIgnoreCase("evaluation"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasSolutions = false;
		boolean hasProblems = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if (arg.equalsIgnoreCase("-solutions")) {
					if (hasSolutions) { printUsageError(); return false; }
					m_solutions = true;
					hasSolutions = true;
					unknownArg = false;
				}					

				if (arg.equalsIgnoreCase("-nosolutions")) {
					if (hasSolutions) { printUsageError(); return false; }
					m_solutions = false;
					hasSolutions = true;
					unknownArg = false;
				}					

				if (arg.equalsIgnoreCase("-problems")) {
					if (hasProblems) { printUsageError(); return false; }
					m_problems = true;
					hasProblems = true;
					unknownArg = false;
				}					

				if (arg.equalsIgnoreCase("-noproblems")) {
					if (hasProblems) { printUsageError(); return false; }
					m_problems = false;
					hasProblems = true;
					unknownArg = false;
				}					


				// Evaluations
				if (arg.equalsIgnoreCase("-alllabeled")||arg.equalsIgnoreCase("-labels")) {
					m_evaluations.add("AllLabeled");
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("evaluate") || arg.equalsIgnoreCase("eval")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the evaluation owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
	        	
	        	Evaluation eval = null;
	        	
	        	for (String clazz : m_evaluations) {
	        		if (clazz=="AllLabeled") {
			        	eval = new AllLabeled(ontology, null, null);
	        		}
		        	
		        	printer.putline("=== " + eval.evaluationClassName() + " ===");
		        	printer.putline(eval.evaluationClassDescription());
		        	printer.putline("Evaluated ontology: " + ontology.getOntologyURI());
		        	printer.putline("Result: " + eval.getEvaluation());
	        		printer.putline("");
	        		if ((m_problems)&&(!eval.getProblems().isEmpty())) {
		        		printer.putline("Problems:");
			        	for (Problem problem : eval.getProblems()) {
			        		printer.putline(problem.getDescription());
			        		printer.putline("Severity: " + problem.getSeverity());
			        		if ((m_solutions)&&(!problem.getSolutions().isEmpty())) {
				        		printer.putline("Possible solutions:");
				        		for (Solution solution : problem.getSolutions()) {
				        			printer.putline(solution.getDescription());
				        			printer.putline("Preference: " + solution.getPreference());
				        		}
			        		}
			        		printer.putline("");
			        	}
	        		}
	        	}
	        	
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			} 
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl evaluate <input_url> [parameters]");
		System.out.println("");
		System.out.println("Evaluates the ontology with the giving evaluation measure");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
//		System.out.println(" -verbose or simply -verb");     
//		System.out.println("     If this switch is set, you will get additional output while processing.");
//		System.out.println("");
	}	
}
