package edu.unika.aifb.owltools.tools;

import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import edu.unika.aifb.owltools.serializer.*;

/**
 * Latex makes a DL-rendering fo an ontology that can be used within Latex
 */
public class Latex extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "latex"; }

	/**
	 * Kind of latex layout
	 */
	enum Layout {
		/**
		 * the simple layout just writes the axioms down without looking for
		 * intendation
		 */
		simple,
		/**
		 * creates a simple table, so the intedation of the axioms looks nice
		 */
		tabular,
		/**
		 * is needed if there are more axioms than can fit on one page. Sometimes
		 * the package is not installed, that's why the normal tabular is available
		 * too
		 */
		longtable,
		/**
		 * Aligns the axioms not with tables, but with other latex hacks
		 */
		aligned
	};
	
	/**
	 * The URI of the ontology to be weakened
	 */
	private String m_fileuri = "";
	
	/**
	 * filename of the output file
	 */
	private String m_output = "";
	
	/**
	 * if complete is turned on, then the output is including the necessary
	 * latex prolog and epilog as well
	 */
	private boolean m_complete = true;
	
	/**
	 * if title is switched on, the ontology name will be printed as well
	 */
	private boolean m_title = true; 
	
	/**
	 * is a very simple layout used?
	 */
	private boolean m_simple = false;
	
	/**
	 * if switched on, default definitions for the new latex commands will be added
	 */
	private boolean m_def = true;
	
	/**
	 * defines the layout to use for the latex output
	 */
	private Layout m_layout = Layout.longtable;
	
	/**
	 * decides if syntax highlighting is to be used or not
	 */
	private boolean m_syntax = true;
	
	/**
	 * should annotations be processed? Can become problematic
	 * with longer comments.
	 */
	private boolean m_annotations = false;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Latex latex = new Latex();
		if (!latex.processArguments(args)) return;
		latex.compile();
		latex.printer.close();
	}

	/**
	 * Processes the command line arguments
	 * @param args The command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("latex"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasTitle = false;
		boolean hasComplete = false;
		boolean hasLayout = false;
		boolean hasSyntax = false;
		boolean hasAnnotations = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				
				if (arg.equalsIgnoreCase("-complete") || arg.equalsIgnoreCase("-c")) {
					if (hasComplete) { printUsageError(); return false; }
					m_complete = true;
					hasComplete = true;
					unknownArg = false;
				}					
				if (arg.equalsIgnoreCase("-fragment") || arg.equalsIgnoreCase("-f")) {
					if (hasComplete) { printUsageError(); return false; }
					m_complete = false;
					hasComplete = true;
					unknownArg = false;
				}					
				
				if (arg.equalsIgnoreCase("-title")) {
					if (hasTitle) { printUsageError(); return false; }
					m_title = true;
					hasTitle = true;
					unknownArg = false;
				}					
				
				if (arg.equalsIgnoreCase("-notitle")) {
					if (hasTitle) { printUsageError(); return false; }
					m_title = false;
					hasTitle = true;
					unknownArg = false;
				}					
				
				if (arg.equalsIgnoreCase("-simple")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = true;
					m_layout = Layout.simple;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-tabular")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = false;
					m_layout = Layout.tabular;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-shorttable")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = false;
					m_layout = Layout.tabular;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-longtable")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = false;
					m_layout = Layout.longtable;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-table")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = false;
					m_layout = Layout.longtable;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-align")) {
					if (hasLayout) { printUsageError(); return false; }
					m_simple = false;
					m_layout = Layout.aligned;
					hasLayout = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-nodefs")) {
					if (hasLayout) { printUsageError(); return false; }
					m_def = false;
					m_simple = false;
					hasLayout = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-syntax")) {
					if (hasSyntax) { printUsageError(); return false; }
					m_syntax = true;
					hasSyntax = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-nosyntax")) {
					if (hasSyntax) { printUsageError(); return false; }
					m_syntax = false;
					hasSyntax = true;					
					unknownArg = false;
				}

				if (arg.equalsIgnoreCase("-ann") || arg.equalsIgnoreCase("-annotations")) {
					if (hasAnnotations) { printUsageError(); return false; }
					m_annotations = true;
					hasAnnotations = true;					
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noann") || arg.equalsIgnoreCase("-noannotations")) {
					if (hasAnnotations) { printUsageError(); return false; }
					m_annotations = false;
					hasAnnotations = true;					
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("latex")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}
	
	/**
	 * creates a latex program out of an ontology
	 * @param ontology The ontology to be tex'ed
	 * @param simple Simple layout or not?
	 * @return the latex program (a list of strings that are the lines in the program)
	 * @throws KAON2Exception
	 */
	public List<String> latex(Ontology ontology, boolean simple) throws KAON2Exception {
		List<String> retval = new LinkedList<String>();
		Request<Axiom> request = ontology.createAxiomRequest();
		
		LatexSerializer latex = new LatexSerializer();
		latex.setSimple(simple);
		
		Set<Axiom> axioms = request.get();
		for (Axiom axiom : axioms) {
			String line = (String)axiom.accept(latex);
			if (line!=null)	if (line.length()>0) retval.add(line + (simple?" \\\\":""));
		}

		return retval;
	}

	/**
	 * Entry point for the command line based usage of the latex owl tool
	 *
	 */
	public void compile() {
		try {
			if (!m_output.equals("")) printer.setFile(m_output);
			Ontology ontology = loadOntology(m_fileuri);
			Filter f = new Filter();
			if (!m_annotations)
				ontology = f.filterOut(ontology, Annotation.class);

	        try {
        		List<String> messages = new LinkedList<String>();
        		
        		if (m_complete) {
        			messages.add("\\documentclass{article}");
        			if (m_layout == Layout.longtable) messages.add("\\usepackage{longtable}");
        			messages.add("\\begin{document}");
        			if (m_title) messages.add("\\section*{" + m_fileuri + "}");
        		}
        		
        		if (m_def) {
        			if ((m_layout == Layout.tabular) || (m_layout == Layout.longtable)) {
    					messages.add("\\newcommand{\\axiom}[3]{$#1$ & $#2$ & $#3$ \\\\}");
    					messages.add("\\newcommand{\\fact}[1]{\\multicolumn{3}{l}{$#1$} \\\\}");
        			}
        			if (m_layout == Layout.aligned) {
    					messages.add("\\newcommand{\\axiom}[3]{$#1$ & $#2$ & $#3$ \\\\}");
    					messages.add("\\newcommand{\\fact}[1]{\\multicolumn{3}{l}{$#1$} \\\\}");
        			}
            		if (m_syntax) {
    					messages.add("\\newcommand{\\concept}[1]{\\textsf{\\textsc{#1}}}");
    					messages.add("\\newcommand{\\property}[1]{\\textsf{#1}}");
    					messages.add("\\newcommand{\\individual}[1]{\\textit{#1}}");        			
    					messages.add("\\newcommand{\\variable}[1]{\\textit{#1}}");        			
            		} else {
    					messages.add("\\newcommand{\\concept}[1]{\\textsf{#1}}");
    					messages.add("\\newcommand{\\property}[1]{\\textsf{#1}}");
    					messages.add("\\newcommand{\\individual}[1]{\\textsf{#1}}");        			
    					messages.add("\\newcommand{\\variable}[1]{\\textsf{#1}}");        			
            		}
        		}        		

        		if (m_layout == Layout.tabular)
    				messages.add("\\begin{tabular}{r@{$\\;\\;$}c@{$\\;\\;$}l}");
        		if (m_layout == Layout.longtable)
    				messages.add("\\begin{longtable}{r@{$\\;\\;$}c@{$\\;\\;$}l}");
        		if (m_layout == Layout.aligned)
    				messages.add("\\begin{tabular}{r@{$\\;\\;$}c@{$\\;\\;$}l}"); // TODO get align from pascal
        		if (m_layout == Layout.simple)
        			messages.add("\\begin{math}");

        		messages.addAll(latex(ontology, m_simple));

        		if (m_layout == Layout.tabular)
    				messages.add("\\end{tabular}");
        		if (m_layout == Layout.longtable)
    				messages.add("\\end{longtable}");
        		if (m_layout == Layout.aligned)
    				messages.add("\\end{tabular}"); // TODO keep up with upper todo
        		if (m_layout == Layout.simple)
        			messages.add("\\end{math}");
        		
        		if (m_complete) {
        			messages.add("\\end{document}");
        		}
    			messages.add("%% created automatically from " + m_fileuri + " by the owl latex tool, see http://owltools.ontoware.org");

        		for (String message : messages )
        			printer.putline(message);
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		} catch (InterruptedException e) {
			printer.error("Interrupted opening KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl latex <input_url> [<output_filename>]");
		System.out.println("latex reads an OWL ontology and returns a latex processable output,");
		System.out.println("that is based on the usual terse DL notation.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the latex code will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println("");
		System.out.println(" -complete    (-c) adds latex-prolog/epilog to create self-contained file");
		System.out.println(" -fragment    (-f) creates a file that can be included in another tex-file");
		System.out.println(" -title       puts in the name of the ontology (default and only when -c)");
		System.out.println(" -notitle     leaves out the name of the ontology");
		System.out.println(" -annotations (short: -ann) also adds the annotations");
		System.out.println(" -noannotations removes annotations prior serialisation (default)");
		System.out.println("");
		System.out.println(" -nodefs      no definitions of the new latex-commands (diy layout))");
		System.out.println(" -simple      a very simple layout");
		System.out.println(" -shorttable  sets a tabular layout (only for small ontologies)");
		System.out.println(" -table       sets a tabular layour, works also for big ontologies (default)");
		//System.out.println(" -align       sets a simple aligned layout");
		System.out.println(" -syntax      concepts, properties and individuals look different (default)");
		System.out.println(" -nosyntax    concepts, properties and individuals look the same");
	}
}
