package edu.unika.aifb.owltools.tools;

import java.util.LinkedList;
import java.util.List;

import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.Ontology;

import edu.unika.aifb.owltools.metrics.*;

/**
 * Accesses several metrics implementations in the metrics package for
 * the command line
 */
public class Metrics extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "metrics"; }
	
	/**
	 * The URI of the ontology to be measured
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";

	/**
	 * list of metrics to be taken
	 */
	private List<String> m_metrics = new LinkedList<String>(); 

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Metrics metrics = new Metrics();
		if (!metrics.processArguments(args)) return;
		metrics.compile();
		metrics.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("metrics")) || (s.equalsIgnoreCase("measure"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				// Metrics
				if (arg.equalsIgnoreCase("-maximaldepth")||arg.equalsIgnoreCase("-md")||arg.equalsIgnoreCase("-m3")) {
					m_metrics.add("MaximalDepth");
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-absolutedepth")||arg.equalsIgnoreCase("-ad")||arg.equalsIgnoreCase("-m1")) {
					m_metrics.add("AbsoluteDepth");
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("metrics") || arg.equalsIgnoreCase("measure")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the evaluation owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
	        	
	        	Metric m = null;
	        	
	        	for (String metric : m_metrics) {
	        		if (metric=="MaximalDepth") {
			        	m = new MaximalDepth(ontology, null, null);
	        		} else if (metric=="AbsoluteDepth") {
	        			m = new AbsoluteDepth(ontology, null, null);
	        		}
		        	
		        	printer.putline(m.metricClassName() + ": " + m.getResult());
		        	printer.putline(m.metricClassDescription());
	        	}
	        	
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			} 
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl metrics <input_url> [parameters]");
		System.out.println("");
		System.out.println("Measures an ontology by a given metric");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println("Add, as parameters, the metrics you want to be measured. The following metrics");
		System.out.println("are available: MaximalDepth");
//		System.out.println(" -verbose or simply -verb");     
//		System.out.println("     If this switch is set, you will get additional output while processing.");
//		System.out.println("");
	}	
}
