/**
 * Populate creates randomly instances for a given TBox
 */
package edu.unika.aifb.owltools.tools;

import java.io.IOException;
import java.util.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.formatting.*;

/**
 * creates instances for classes
 */
public class Populate extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "populate"; }

	/**
	 * The URI of the ontology to be filtered
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to save the ontology
	 */
	private OntologyFileFormat m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, the result will be supressed 
	 */
	private boolean m_result = true;
	
	/**
	 * number of instances to create
	 */
	private int m_number = 0;
	
	/**
	 * Random variable for the name generator
	 */
	private Random m_random = new Random();
	
	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Populate pop = new Populate();
		if (!pop.processArguments(args)) return;
		pop.compile();
		pop.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("pop")) || (s.equalsIgnoreCase("populate"))) {
				printUsage();
				return false;
			}
		}
		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];
			
			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				if (java.lang.Character.isDigit(arg.charAt(1))) {
					try {
						m_number = java.lang.Integer.valueOf(arg.substring(1));
					} catch (NumberFormatException n) {
						printUsageError();
						return false;
					}
					unknownArg = false;
				}

				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.ABSTRACT_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if ((arg.equalsIgnoreCase("pop")) || (arg.equalsIgnoreCase("populate"))) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}

	/**
	 * A very sophisticated algorithm to create new URIs for the individuals
	 * to populate the ontology with
	 * @param random
	 * @return
	 */
	String newURI() {
		return "http://www.example.com/instance#i" + Math.abs(m_random.nextInt());
	}
	
	/**
	 * Populates an ontology with newly created individuals. The individuals are
	 * simply sticked to different classes, randomly.
	 * @param ontology The ontology to be populated
	 * @param instances The number of instances to generate
	 * @return The populated ontology
	 * @throws KAON2Exception
	 */
	public Ontology populate(Ontology ontology, int instances) throws KAON2Exception {
		Request<OWLClass> request = ontology.createEntityRequest(OWLClass.class);
		int size = request.size();
		OWLClass[] classes = new OWLClass[size];
		Set<OWLClass> classset = request.get();
		int i = 0;
		for (OWLClass clazz : classset)
			classes[i++] = clazz;
		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		for (i = 0; i < instances; i++) {
			int rnd = m_random.nextInt(size);
			Individual instance = KAON2Manager.factory().individual(newURI());
			ClassMember classmembership = KAON2Manager.factory().classMember(classes[rnd], instance);
			printer.verbose("Adding " + classmembership.toString());
			changes.add(new OntologyChangeEvent(classmembership, OntologyChangeEvent.ChangeType.ADD));
		}
		ontology.applyChanges(changes);
		return ontology;
	}

	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {

	        	Ontology ontology = loadOntology(m_fileuri);
	        	ontology = populate(ontology, m_number);
	        	
	        	if (m_result)
	        		ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl populate <input_url> <output_filename>");
		System.out.println("");
		System.out.println("Populates an ontology randomly with newly created instances");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be populated.");
		System.out.println("");
		System.out.println(" <output_filename>");
		System.out.println("     The filename of the file, where the populated ontology will be written to.");
		System.out.println("     If omitted, the output will be written to the command line.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result. In combination with -verbose you");
		System.out.println("     get a list of the needed changes, without apllying them.");
	}
	
	
}
