/*
 * Creates a CSV file out of an ontology,so that it can be read into
 * spreadsheet tools.
 */
package edu.unika.aifb.owltools.tools;

import java.util.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.logic.*;

/**
 * Reads an ontology and creates a CSV file out of it.
 * 
 * @author Denny Vrandecic
 */
public class CSV extends KAON2Tool {
	/**
	 * Name of the tool
	 */
	@Override
	public String name() { return "csv"; }
	
	/**
	 * Input filename
	 */
	private String m_fileuri = "";
	
	/**
	 * Name of the output file
	 */
	private String m_output = "";

	/**
	 * If the tool is being called from the outside, some basic processing
	 * is being done
	 * @param args Command line parameters. The same as for the OWLtools call
	 */
	public static void main(String[] args) {
		CSV csv = new CSV();
		if (!csv.processArguments(args)) return;
		csv.compile();
		csv.printer.close();
	}

	/**
	 * processes the command line parameteres
	 */
	@Override
	public boolean processArguments(String[] args) {
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("dump")) || (s.equalsIgnoreCase("count"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
								
			} else {
				if (arg.equalsIgnoreCase("csv")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	private static String name(Entity e) {
		String result = e.getURI();
		if (result.lastIndexOf('#') > 0) {
			result = result.substring(result.lastIndexOf('#') + 1);
		} else if (result.lastIndexOf("/") > 0) {
			result = result.substring(result.lastIndexOf('/') + 1);
		}
		return result;
	}
	
	private static String value(Object o) {
		String result = o.toString();
		if (Constant.class.isInstance(o)) {
			Constant val = (Constant)o;
			result = val.getValue().toString();
		}
		return result;
	}
	
	private static String semicola(int i) {
		String result = "";
		for (int j = 0; j < i; j++) { result += ";"; }
		return result;
	}

	public static List<String> csv(Ontology ontology) throws KAON2Exception, InterruptedException {
		List<String> result = new LinkedList<String>();
		Reasoner reasoner = ontology.createReasoner();
		
		Map<DataProperty, Map<Individual, Set<Object>>> values = new HashMap<DataProperty, Map<Individual, Set<Object>>>();
		Map<DataProperty, Integer> length = new HashMap<DataProperty, Integer>(); 
		
		Request<DataProperty> dprequest = ontology.createEntityRequest(DataProperty.class);
		Set<DataProperty> dps = dprequest.get();
		String line = "";
		for (DataProperty dp : dps) {
			length.put(dp, 0);
			Query query = reasoner.createQuery(dp);
			query.open();
			Map<Individual, Set<Object>> dpvalues = new HashMap<Individual, Set<Object>>();
			while (!query.afterLast()) {
				query.next();
				Object[] t = query.tupleBuffer();
				if (!(t.length==2)) {
					System.out.println("Wrong length");
					System.exit(0);
				}
				if (!(Individual.class.isInstance(t[0]))) {
					System.out.println("Wrong type");
					System.exit(0);
				}
				Individual i = (Individual)t[0];
				if (dpvalues.containsKey(i)) {
					Set<Object> v = dpvalues.get(i);
					v.add(t[1]);
					dpvalues.put(i, v);
					Integer maxlen = v.size();
					if (length.get(dp) < maxlen)
						length.put(dp, maxlen);
				} else {
					Set<Object> v = new HashSet<Object>();
					v.add(t[1]);
					dpvalues.put(i, v);
					if (length.get(dp) == 0)
						length.put(dp, 1);
				}
				values.put(dp, dpvalues);
			}
			if (length.get(dp) > 0) {
				line += ";" + name(dp);
				line += semicola(length.get(dp) - 1);
			}
			query.close();
		}
		result.add(line);
		Request<Individual> irequest = ontology.createEntityRequest(Individual.class);
		Set<Individual> individuals = irequest.get();
		for (Individual individual : individuals) {
			line = name(individual);
			for (DataProperty dpi : dps) {
				Map<Individual, Set<Object>> dpvalues = values.get(dpi);
				Set<Object> v = dpvalues.get(individual);
				if (v == null) {
					line += semicola(length.get(dpi));
				} else if (v.size() == 0) {
					line += semicola(length.get(dpi));
				} else if (v.size() == 1) {
					line += ";" + value(v.toArray()[0]);
					line += semicola(length.get(dpi)-1);
				} else {
					line += ";";
					for (Object object : v) {
						line += value(object) + ";";
					}
					line = line.substring(0, line.length()-1);
					line += semicola(length.get(dpi)-v.size());
				}
			}
			result.add(line);
		}	
		
		return result;
	}
	
	/**
	 * Entry point for the command line based usage of the Dump owl tool
	 *
	 */
	public void compile() {
		try {
			if (!m_output.equals("")) printer.setFile(m_output);
			Ontology ontology = loadOntology(m_fileuri);

			List<String> lines = csv(ontology);
    		for (String line : lines )
    			printer.putline(line);

			endConnection();
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		} catch (InterruptedException e) {
			printer.error("Interrupted opening KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl csv <input_url> [<output_filename>]");
		System.out.println("csv reads an OWL ontology and creates an CSV file.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the CSV file to write.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
	}
	
	
}
