package edu.unika.aifb.owltools.tools;

import java.util.*;
import java.io.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.formatting.*;

import edu.unika.aifb.owltools.serializer.*;
import edu.unika.aifb.owltools.serializer.ProgramSerializer.*;
import edu.unika.aifb.owltools.utils.*;
import edu.unika.aifb.owltools.api;

/**
 * DLPconvert is a tool to transform different syntactical representations of
 * a DLP program into each other (refer to Raphael Volz, 2004). This is the main
 * class that calls the actual transformation algorithms provided in KAON2 and in
 * the other classes of this package.
 */
public class DLPConvert extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "dlpconvert"; }

	/**
	 * Specifies the different possible formats that dlpconvert is able to serialise
	 * a set of rules in.
	 */
	public enum Format {
		/**
		 * Disjunctive datalog. Like Prolog, but may still contain disjunctions in the head. Default.
		 */
		DDatalog,
		/**
		 * The format directly usable by prolog implementations, especially by XSB Prolog.
		 */
		Prolog,
		/**
		 * The RuleML format
		 */
		RuleML,
		/**
		 * The KAON2 Rules format
		 */
		Rule,
		/**
		 * FLogic, can be directly used by OntoBroker and other FLogic-engines.
		 * This format encodes every literal via F-Logic molecules and class hierarchies.
		 */
		FLogic,
		/**
		 * FLogic, can be directly used by OntoBroker and other FLogic-engines.
		 * This format returns code that is very similar to Prolog, but with additional
		 * explicit quantifiers.
		 */
		LPFLogic,
		/**
		 * Proposed SWRL extension to OWL
		 */
		SWRL
	};
	
	/**
	 * Possibilities of how to handle integrity constraints. Integrity constraints are
	 * rules with an empty head, that never must evaluate to true. They are not natively
	 * supported by many logic programming environments, but can be emulated with an
	 * auxilliery goal.
	 */
	public enum IntegrityConstraintHandling {
		/**
		 * Integrity constraints are rendered as rules with an empty head. Some
		 * implementations cannot deal with such a result
		 */
		asIC,
		/**
		 * Integrity constraints are simply skipped. The result is weaker than the input
		 */
		noIC,
		/**
		 * A new predicate is included. After the evaluation of the results, it can be
		 * checked if the predicate is true. If so, an integrity constraint was violated.
		 */
		ICpredicate
	}
	
	/**
	 * The format for the output.
	 */
	private Format m_format = Format.DDatalog;
	
	/**
	 * Should the ABox be transformed as well?
	 */
	private boolean m_aBox = true;
	
	/**
	 * Should the programm be optimised for stand-alone use. This
	 * means that rules that cannot contribute to further inferences
	 * are dropped, and that some rules might be simplified.
	 */
	private boolean m_single = false;
	
	/**
	 * Should the rule base be unfolded? This is a KAON2 internal thing
	 */
	private boolean m_folding = true;
	
	/**
	 * Do we allow for disjunctive heads in the rules? Note that some serializations
	 * actually treat several heads as conjunctive heads. The output here is meant
	 * to be disjunctive and thus, if this is set on, some outputs are not conformant
	 * to standards.
	 */
	private boolean m_allowDDL = false;
	
	/**
	 * Should disjunctive heads be transformed to conjunctie ones? This leads to
	 * changed semantics of the program and usually better performance. Especially,
	 * integrity constraints can be dropped without loosing results, but additional
	 * wrong results can be inferred.
	 * Refer to Hitzler, Vrandecic at the ISWC2005
	 */
	private boolean m_screech = false;
	
	/**
	 * If switched on, some statistics during the screech process are counted.
	 */
	private boolean m_stat = false;
	
	/**
	 * How should names be serialized? Most implementations cannot deal with URIs, so
	 * a few different serizalization styles are included.
	 */
	private NameSerializationStyle m_nameStyle = NameSerializationStyle.uri;
	
	/**
	 * The URI of the ontology to be converted
	 */
	private String m_fileuri = "";

	/**
	 * The filename for the output to be saved
	 */
	private String m_output = "";

	/**
	 * How to handle integrity constraints
	 */
	private IntegrityConstraintHandling m_ichandling = IntegrityConstraintHandling.asIC; 
	
	/**
	 * This is used for the Tomcat based implementation, to count the usage of the tool
	 */
	private static int count = 0;

	/**
	 * Allows to start dlpconvert from the command line as well.
	 * Only used for a command line version.
	 * 
	 * @param args command line parameters
	 */
	public static void main(String[] args) {
		DLPConvert dlp = new DLPConvert();
		if (!dlp.processArguments(args)) return;
		dlp.compile();
		dlp.printer.close();
	}
	
	/**
	 * Processes the command line parameters and sets the appropriate flags instead.
	 * Only used for the command line version.
	 * 
	 * @param args command line parameters
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("dlpconvert")) || (s.equalsIgnoreCase("dlp")) || (s.equalsIgnoreCase("screech"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasFormat = false;
		boolean hasNameStyle = false;
		boolean hasABox = false;
		boolean hasFolding = false;
		boolean hasDDL = false;
		boolean hasVerbose = false;
		boolean hasIC = false;

		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;

			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {

				// Format
				if (arg.equalsIgnoreCase("-datalog")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.DDatalog;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-Prolog")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.Prolog;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-KAON2")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.Rule;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-RuleML")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.RuleML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-FLogic")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.FLogic;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-LPFLogic")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.LPFLogic;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SWRL")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = Format.SWRL;
					hasFormat = true;
					unknownArg = false;
				}

				// Name serialization style
				if (arg.equalsIgnoreCase("-URI")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.uri;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-atom")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.atom;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-short")) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.fragment;
					hasNameStyle = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ns") || (arg.equalsIgnoreCase("-Namespace")) ) {
					if (hasNameStyle) { printUsageError(); return false; }
					m_nameStyle = NameSerializationStyle.ns;
					hasNameStyle = true;
					unknownArg = false;
				}
				
				// to ABox or not to ABox
				if (arg.equalsIgnoreCase("-abox")) {
					if (hasABox) { printUsageError(); return false; }
					m_aBox = true;
					hasABox = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noabox")) {
					if (hasABox) { printUsageError(); return false; }
					m_aBox = false;
					hasABox = true;
					unknownArg = false;
				}

				// remove unfounded rules
				if (arg.equalsIgnoreCase("-single")) {
					m_single = true;
					unknownArg = false;
				}

				// folding
				if (arg.equalsIgnoreCase("-fold") || arg.equalsIgnoreCase("-unfold")) {
					if (hasFolding) { printUsageError(); return false; }
					m_folding = true;
					hasFolding = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-dontfold") || arg.equalsIgnoreCase("-nounfold")) {
					if (hasFolding) { printUsageError(); return false; }
					m_folding = false;
					hasFolding = true;
					unknownArg = false;
				}

				// DDL
				if (arg.equalsIgnoreCase("-ddl") || arg.equalsIgnoreCase("-full") || arg.equalsIgnoreCase("-x") || arg.equalsIgnoreCase("-xdlp")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = true;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noddl") || arg.equalsIgnoreCase("-dlp") || arg.equalsIgnoreCase("-noxdlp")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = false;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-screech")) {
					if (hasDDL) { printUsageError(); return false; }
					m_allowDDL = true;
					m_screech = true;
					hasDDL = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-screechcount") || arg.equalsIgnoreCase("-sc")) {
					if (hasDDL) { printUsageError(); return false; }
					m_stat = true;
					m_allowDDL = true;
					m_screech = true;
					hasDDL = true;
					unknownArg = false;
				}
				
				// Integrity Constraint handling
				if (arg.equalsIgnoreCase("-ic")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.asIC;
					hasIC = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-noic")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.noIC;
					hasIC = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-icpredicate") || arg.equalsIgnoreCase("-icp")) {
					if (hasIC) { printUsageError(); return false; }
					m_ichandling = IntegrityConstraintHandling.ICpredicate;
					hasIC = true;
					unknownArg = false;
				}
				// Prolog and FLogic require some IC handling. Use ICP if none given.
				if ( (m_ichandling == IntegrityConstraintHandling.asIC) && ( (m_format == Format.Prolog) || (m_format == Format.FLogic) ) ) {
					m_ichandling = IntegrityConstraintHandling.ICpredicate;
				}
				
				if (arg.equalsIgnoreCase("-verb") || arg.equalsIgnoreCase("-verbose")) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}

			} else {
				if (arg.equalsIgnoreCase("dlpconvert") || arg.equalsIgnoreCase("dlp")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;
						unknownArg = false;
					}
				}
			}

			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Creates an ontology consisting of the rules that KAON2's conversion to disjunctive
	 * datalog creates out of the given ontology.
	 * 
	 * @param in The ontology to be created.
	 * @param out The ontology that will hold the resulting rulebase.
	 * @throws KAON2Exception
	 * @throws InterruptedException
	 */
	public void dlpconvert(Ontology in, Ontology out) throws KAON2Exception, InterruptedException {
		Clear.clear(out);
		Reasoner reasoner = in.createReasoner();
		Collection<Rule> rules = reasoner.getReductionToDisjunctiveDatalog(true, true, true, true);
		rules = addTransitivity(rules, in);
		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		for (Rule rule : rules)
			changes.add(new OntologyChangeEvent(rule, OntologyChangeEvent.ChangeType.ADD));
		out.applyChanges(changes);
	}

	/**
	 * Convenience interface. If you have a name for your result ontology, you can use this.
	 * 
	 * @param ontology Ontology to be converted
	 * @param uri the URI of the ontology to be created
	 * @param physicalUri where is the ontology physically saved
	 * @return the converted ontology
	 * @throws KAON2Exception from KAON2's reduction
	 * @throws InterruptedException from KAON2's reduction
	 */
	public Ontology dlpconvert(Ontology ontology, String uri, String physicalUri) throws KAON2Exception, InterruptedException {
        DefaultOntologyResolver resolver = new DefaultOntologyResolver();
        resolver.registerReplacement(uri, physicalUri);
        ontology.getConnection().setOntologyResolver(resolver);
		Ontology rv = ontology.getConnection().createOntology(uri, new HashMap<String,Object>());
		dlpconvert(ontology, rv);
		return rv;
	}

	/**
	 * Convenience interface, so you can easily work with ontologies without having to define
	 * a name for it. Useful for testing or for temporary results.
	 * 
	 * @param ontology Ontology to be converted
	 * @return Converted ontology
	 * @throws KAON2Exception from KAON2's reduction
	 * @throws InterruptedException from KAON2's reduction
	 */
	public Ontology dlpconvert(Ontology ontology) throws KAON2Exception, InterruptedException {
		return dlpconvert(ontology, "http://owltools.ontoware.org/dlpconvert/", "file:C:/temp/dlpconvert.xml");
	}

	// public static void dlpconvert(Ontology ontology) <- replaces the very ontology
	// public static void screech(Ontology in, Ontology out)
	// public static Ontology screech(Ontology ontology, String uri, String physicalUri)
	// public static void screech(Ontology ontology) <- replaces inline
	// public static void removeICConstraints(Ontology ontology) removes IC Constraints inline
	// breakDisjunctions(O
	
	//////////////////////////////////////////////////////////////////////////////
	//////////
	//////////  Transformation Methods
	//////////
	//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Apply the given substitution to the given Literal and return the result.
	 */
	protected Literal substitute(Literal l, Map<Variable,Term> s) {
		List<Term> newargs = new ArrayList<Term>();
		Term arg;
		for (int i=0; i<l.getArity(); i++) {
			newargs.add(substitute(l.getArgument(i),s));
		}
		return KAON2Manager.factory().literal(l.isPositive(),l.getPredicate(),newargs);
	}

	/**
	 * Apply the given substitution to the given Term and return the result.
	 */
	protected Term substitute(Term t, Map<Variable,Term> s) {
		if (t instanceof Variable) {
			if (s.containsKey(t)) {
				return s.get(t);
			} else {
				return t;
			}
		} else if (t instanceof FunctionalTerm) {
			List<Term> newargs = new ArrayList<Term>();
			for (int i=0; i<t.getArity(); i++) {
				newargs.add(substitute(t.getArgument(i),s));
			}
			return KAON2Manager.factory().functionalTerm(((FunctionalTerm)t).getFunctionSymbol(), newargs);
		} else {
			return t;
		}
	}

	/**
	 * Add ABox axioms from a given ontology to a collection of rules, and instantiate auxilliary
	 * axioms for those using function symbols. This method also takes care of substituting anonymous
	 * variables whenever variables are singleton.
	 */
	public Collection<Rule> addABox(Collection<Rule> rules, Ontology ontology) throws KAON2Exception {
		Rule rule;
		ArrayList<Rule> result = new ArrayList<Rule>();
		List<Formula> emptybody = new ArrayList<Formula>();
		HashMap<String,Term> individuals = new HashMap<String,Term>(); // encountered terms, use hash to avoid duplicates

		// add all ABox axioms as rules
		Cursor<ClassMember> cmcursor = ontology.createAxiomRequest(ClassMember.class).openCursor();
		while (cmcursor.hasNext()) {
			ClassMember axiom = cmcursor.next();
			rule = KAON2Manager.factory().rule( KAON2Manager.factory().literal(true, axiom.getDescription(), axiom.getIndividual()), emptybody);
			result.add(rule);
			individuals.put(axiom.getIndividual().toString(), axiom.getIndividual());
		}
		Cursor<ObjectPropertyMember> opmcursor = ontology.createAxiomRequest(ObjectPropertyMember.class).openCursor();
		while (opmcursor.hasNext()) {
			ObjectPropertyMember axiom = opmcursor.next();
			rule = KAON2Manager.factory().rule( KAON2Manager.factory().literal(true, axiom.getObjectProperty(), axiom.getSourceIndividual(), axiom.getTargetIndividual()), emptybody);
			result.add(rule);
			individuals.put(axiom.getSourceIndividual().toString(), axiom.getSourceIndividual());
			individuals.put(axiom.getTargetIndividual().toString(), axiom.getTargetIndividual());
		}
		Cursor<DataPropertyMember> dpmcursor = ontology.createAxiomRequest(DataPropertyMember.class).openCursor();
		while (dpmcursor.hasNext()) {
			DataPropertyMember axiom = dpmcursor.next();
			rule = KAON2Manager.factory().rule( KAON2Manager.factory().literal(true, axiom.getDataProperty(), axiom.getSourceIndividual(), axiom.getTargetValue()), emptybody);
			result.add(rule);
			individuals.put(axiom.getSourceIndividual().toString(), axiom.getSourceIndividual());
			individuals.put(axiom.getTargetValue().toString(), axiom.getTargetValue());
		}

		// instantiate universe
		PredicateSymbol kaon2O = KAON2Manager.factory().predicateSymbol(PredicateSymbol.NAMED_OBJECTS_URI,1);
		for (Term term : individuals.values()) {
			rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true,kaon2O,term), emptybody);
			result.add(rule);
		}
		individuals.clear();

		// find relevant function symbols
		String uri, fname;
		HashSet<String> functions = new HashSet<String>(); // encountered function symbols
		HashSet<Predicate> predicates = new HashSet<Predicate>(); // encountered predicate symbols
		HashSet<Variable> singlevars = new HashSet<Variable>(); // singleton variables
		HashSet<Variable> nonsinglevars = new HashSet<Variable>(); // non-singleton variables
		HashMap<Variable,Term> varsubst = new HashMap<Variable,Term>(); // replacements for variables
		// Rule does not allow us to check the URI of its Literals' Predicates, since KAON2
		// allows predicates without URIs. Casts are needed. Argh.
		for (Rule r : rules) {
			varsubst.clear();
			singlevars.clear();
			nonsinglevars.clear();
			// record variables in head:
			List<Literal> lits = Arrays.asList(r.getHeadLiterals());
			for (Literal l : lits) {
				predicates.add(l.getPredicate());
				for (int i=0; i<l.getArity(); i++) {
					for (Variable v : l.getArgument(i).getFreeVariables()) {
						if (!nonsinglevars.contains(v)) {
							if (singlevars.contains(v)) {
								nonsinglevars.add(v);
								singlevars.remove(v);
							} else {
								singlevars.add(v);
							}
						}
					}
				}
			}

			lits = Arrays.asList(r.getBodyLiterals());
			// first iteration: compute required variable substitutions, if any
			// and record variables in body
			for (Literal l : lits) {
				Predicate p = l.getPredicate();
				predicates.add(p);
				if ( (p instanceof PredicateSymbol) && (l.getArity() == 2) ) {
					uri = ((PredicateSymbol)p).getURI();
					fname = "";
					if (uri.startsWith(PredicateSymbol.ABSTRACT_SUCCESSOR_PREDICATE_URI_PREFIX)) {
						fname = "kaon2afunc_" + uri.substring(PredicateSymbol.ABSTRACT_SUCCESSOR_PREDICATE_URI_PREFIX.length());
						functions.add(fname);
					} else if (uri.startsWith(PredicateSymbol.CONCRETE_SUCCESSOR_PREDICATE_URI_PREFIX)) {
						fname = "kaon2cfunc_" + uri.substring(PredicateSymbol.CONCRETE_SUCCESSOR_PREDICATE_URI_PREFIX.length());
						functions.add(fname);
					}
					if ( fname != "" ) {
						Term x = l.getArgument(0);
						Term fx = KAON2Manager.factory().functionalTerm(fname, x);
						if (l.getArgument(1) instanceof Variable) {
							Variable xf = (Variable)l.getArgument(1);
							varsubst.put(xf,fx);
						}
					}
				}
				// record variables in body
				for (int i=0; i<l.getArity(); i++) {
					for (Variable v : l.getArgument(i).getFreeVariables()) {
						if (!nonsinglevars.contains(v)) {
							if (singlevars.contains(v)) {
								nonsinglevars.add(v);
								singlevars.remove(v);
							} else {
								singlevars.add(v);
							}
						}
					}
				}
			}

			if (m_format == Format.Prolog) { // F-Logic knows no anonymous variables
				for (Variable v : singlevars) {
					if (!varsubst.containsKey(v)) {
						varsubst.put(v, KAON2Manager.factory().variable("_" + v.getVariableName()));
					}
				}
			}

			// second iteration: make function symbols explicit again
			if (varsubst.size() > 0) {
				// Grmph. Java-style array concatenation:
				//lits = new ArrayList<Literal>(Arrays.asList(r.getHeadLiterals()));
				//lits.addAll(Arrays.asList(r.getBodyLiterals()));
				List<Literal> head = new ArrayList<Literal>();
				List<Literal> body = new ArrayList<Literal>();
				HashSet<Term> safevars = new HashSet<Term>(); // variables that are already safe due to added kaon2namedobject-statements
				for (Literal l : lits) {
					Predicate p = l.getPredicate();
					if (p instanceof PredicateSymbol) {
						uri = ((PredicateSymbol)p).getURI();
						if ( (uri.startsWith(PredicateSymbol.ABSTRACT_SUCCESSOR_PREDICATE_URI_PREFIX)) || (uri.startsWith(PredicateSymbol.CONCRETE_SUCCESSOR_PREDICATE_URI_PREFIX)) ) {
							// add literal to ground x in named objects
							Term x = l.getArgument(0);
							if (!safevars.contains(x)) {
								body.add(KAON2Manager.factory().literal(true, kaon2O, x));
								safevars.add(x);
							}
						} else {
							body.add(substitute(l,varsubst));
						}
					} else body.add(substitute(l,varsubst)); // can this happen?
				}
				lits = Arrays.asList(r.getHeadLiterals());
				for (Literal l : lits) {
					head.add(substitute(l,varsubst));
				}
				rule = KAON2Manager.factory().rule(head, r.isHeadConjunctive(), body);
				result.add(rule);
			} else {
				result.add(r);
			}
		}

		// Herbrand universe consists of base individuals and their function terms of depth 1
		Term x = KAON2Manager.factory().variable("X");
		PredicateSymbol kaon2HU = KAON2Manager.factory().predicateSymbol(PredicateSymbol.HU_URI,1);
		List<Formula> Obody = new ArrayList<Formula>();
		Obody.add(KAON2Manager.factory().literal(true, kaon2O, x));
		rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true, kaon2HU, x), Obody);
		result.add(rule);

		for (String s : functions) {
			rule = KAON2Manager.factory().rule( 
				KAON2Manager.factory().literal(true, kaon2HU, KAON2Manager.factory().functionalTerm(s, x)), 
				Obody
			);
			result.add(rule);
		}
		functions.clear();

		// add equality theory
		PredicateSymbol kaon2equal = KAON2Manager.factory().predicateSymbol(PredicateSymbol.EQUAL_URI,2);
		Term y = KAON2Manager.factory().variable("Y");
		Term z = KAON2Manager.factory().variable("Z");
		List<Formula> refbody = new ArrayList<Formula>();
		refbody.add(KAON2Manager.factory().literal(true, kaon2HU, x));
		List<Formula> symbody = new ArrayList<Formula>();
		symbody.add(KAON2Manager.factory().literal(true, kaon2HU, x));
		symbody.add(KAON2Manager.factory().literal(true, kaon2HU, y));
		symbody.add(KAON2Manager.factory().literal(true, kaon2equal, y, x));
		List<Formula> transbody = new ArrayList<Formula>();
		transbody.add(KAON2Manager.factory().literal(true, kaon2HU, x));
		transbody.add(KAON2Manager.factory().literal(true, kaon2HU, y));
		transbody.add(KAON2Manager.factory().literal(true, kaon2HU, z));
		transbody.add(KAON2Manager.factory().literal(true, kaon2equal, x, y));
		transbody.add(KAON2Manager.factory().literal(true, kaon2equal, y, z));
		rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true, kaon2equal, x, x), refbody);
		result.add(rule);
		rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true, kaon2equal, x, y), symbody);
		result.add(rule);
		rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true, kaon2equal, x, z), transbody);
		result.add(rule);
		// add equality congruences
		List<Term> args1 = new ArrayList<Term>();
		List<Term> args2 = new ArrayList<Term>();
		List<Formula> body = new ArrayList<Formula>();
		boolean skip;
		for (Predicate p : predicates) {
			skip = false;
			if (p instanceof PredicateSymbol) {
				uri = ((PredicateSymbol)p).getURI();
				if ( (uri.startsWith(PredicateSymbol.ABSTRACT_SUCCESSOR_PREDICATE_URI_PREFIX)) || (uri.startsWith(PredicateSymbol.CONCRETE_SUCCESSOR_PREDICATE_URI_PREFIX)) || (uri.startsWith(PredicateSymbol.EQUAL_URI)) ||
				(uri.startsWith(PredicateSymbol.EVALUATE_URI)) || (uri.startsWith(PredicateSymbol.HU_URI)) || (uri.startsWith(PredicateSymbol.IF_TRUE_URI)) || (uri.startsWith(PredicateSymbol.NAMED_OBJECTS_URI)) || (uri.startsWith(PredicateSymbol.NOT_EQUAL_URI)) ) {
					skip = true;
				}
			}
			if (!skip) {
				for (int i=0; i<p.getArity(); i++) {
					args1.clear();
					args2.clear();
					body.clear();
					for (int j=0; j<p.getArity(); j++) {
						if (i == j) {
							args1.add(x);
						} else {
							args1.add(KAON2Manager.factory().variable("X_" + j));
						}
					}
					args2.addAll(args1);
					args2.set(i,y);
					body.add(KAON2Manager.factory().literal(true, p, args2));
					body.add(KAON2Manager.factory().literal(true, kaon2equal, x, y));
					rule = KAON2Manager.factory().rule(KAON2Manager.factory().literal(true, p, args1), body);
					result.add(rule);
				}
			}
		}

		return result;
	}

	/**
	 * Explicitly adds transitivity rules to a rule set. This enables some additional queries
	 * involving negated roles that are not simple to be answered correctly.
	 */
	public Collection<Rule> addTransitivity(Collection<Rule> rules, Ontology ontology) throws KAON2Exception {
		Cursor<ObjectPropertyAttribute> cursor = ontology.createAxiomRequest(ObjectPropertyAttribute.class).openCursor();
		while (cursor.hasNext()) {
			ObjectPropertyAttribute opa = cursor.next();
			if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_TRANSITIVE) {
				// build new tranisitivy rule and add it
				Predicate predicate = opa.getObjectProperty();
				Term x = KAON2Manager.factory().variable("X");
				Term y = KAON2Manager.factory().variable("Y");
				Term z = KAON2Manager.factory().variable("Z");
				Term[] xz = new Term[2];
				Term[] xy = new Term[2];
				Term[] yz = new Term[2];
				xz[0] = xy[0] = x;
				xy[1] = yz[0] = y;
				xz[1] = yz[1] = z;
				Literal head = KAON2Manager.factory().literal(true, predicate, xz);
				List<Literal> body = new LinkedList<Literal>();
				body.add(KAON2Manager.factory().literal(true, predicate, xy));
				body.add(KAON2Manager.factory().literal(true, predicate, yz));
				Rule rule = KAON2Manager.factory().rule(head, body);
				rules.add(rule);
			}
		}
		return rules;
	}

	/**
	 * Sort a given collection of Rule objects using a RuleComparator.
	 */
	public List<Rule> sortRules(Collection<Rule> rules) {
		ArrayList<Rule> sortrules = new ArrayList<Rule>(rules);
		Collections.sort(sortrules, new RuleComparator());
		return sortrules;
	}
	
	/**
	 * Remove all unfounded rules. The set of "defined" predicate symbols is
	 * the smallest set that contains all predicate symbols P satisfying the
	 * the following property:
	 *
	 * P occurs in the head of at least one rule the body of which contains 
	 * only defined predicate symbols. Especially this is the case if P occurs 
	 * in a fact.
	 *
	 * A rule is unfounded if its body contains a predicate symbol that is not
	 * defined in the above sense. This method removes all those rules.
	 */
	public Collection<Rule> removeUnfoundedRules(Collection<Rule> rules) {
		ArrayList<Rule> result = new ArrayList<Rule>();
		HashSet<String> defpreds = new HashSet<String>();
		boolean changed = true;
		boolean defined;
		Iterator<Rule> ri;
		Rule rule;
		// do a least fixpoint computation
		while (changed) {
			changed = false;
			ri = rules.iterator();
			while (ri.hasNext()) {
				rule = ri.next();
				defined = true;
				for (int i=0; i<rule.getBodyLength(); i++) {
					if ( !defpreds.contains(rule.getBodyLiteral(i).getPredicate().toString() + "(" + rule.getBodyLiteral(i).getArity()) ) {
						defined = false;
						continue;
					}
				}
				if (defined) {
					for (int i=0; i<rule.getHeadLength(); i++) {
						defpreds.add(rule.getHeadLiteral(i).getPredicate().toString() + "(" + rule.getHeadLiteral(i).getArity());
					}
					result.add(rule);
					ri.remove();
					changed = true;
				}
			}
		}
		return result;
	}

	/**
	 * Transform or delete integrity contraints (rules without positive part/head) in rule 
	 * collections.
	 * 
	 * @param rules the rules to be converted
	 * @return transformed collection of rules
	 */
	public Collection<Rule> transformICs(Collection<Rule> rules) {
		Collection<Rule> result = new LinkedList<Rule>();
		Literal inclit = null;
		if (m_ichandling == IntegrityConstraintHandling.ICpredicate) {
			PredicateSymbol incpred=KAON2Manager.factory().predicateSymbol("http://owltools.ontoware.org/concepts/dlpconvert/inconsistent",0);
			inclit = KAON2Manager.factory().literal(true, incpred);
		}
		for (Rule rule : rules) {
			if (rule.getHeadLength() == 0) {
				switch (m_ichandling) {
					case asIC:
						result.add(rule);
					break;
					case noIC: break;
					case ICpredicate:
						Rule newrule = KAON2Manager.factory().rule(inclit, rule.getBodyLiterals());
						result.add(newrule);
					break;
				}
			} else {
				result.add(rule);
			}
		}
		return result;
	}

	/**
	 * Eliminate disjunctions in rule collections by splitting rules. If there is a disjunction 
	 * in the head, it is changed into a conjunction (Screech method). The result may be applied 
	 * for approximate reasoning.
	 * 
	 * @param rules the rules to be converted
	 * @return the screech program, as a collection of rules
	 */
	public Collection<Rule> screech(Collection<Rule> rules) {
		Collection<Rule> result = new LinkedList<Rule>();
		
		int disjrulecounter = 0;
		int disjcounter = 0; 
		int iccounter = 0;

		for (Rule rule : rules) {
			if (rule.getHeadLength() == 0) { // just drop integrity contraints, this is ok in screech
				iccounter++;
			} else if (rule.getHeadLength() == 1) { // keep definite rules unchanged
				result.add(rule);
			} else { // split disjunctions, might create duplicates
				disjrulecounter++;
				int l = rule.getHeadLength();
				disjcounter += l;
				List<Literal> body = new ArrayList<Literal>();
				int b = rule.getBodyLength();
				for (int i = 0; i < b; i++) {
					body.add(rule.getBodyLiteral(i));
				}
				for (int i = 0; i < l; i++) {
					List<Literal> head = new ArrayList<Literal>();
					head.add(rule.getHeadLiteral(i));
					Rule approxrule = KAON2Manager.factory().rule(head, false, body);
					result.add(approxrule);
				}
			}
		}
		if (m_stat) printer.verbose(disjcounter + " disjunctions");
		if (m_stat) printer.verbose(", " + disjrulecounter + " disjunctive rules");
		if (m_stat) printer.verbose(", " + iccounter + " ICs");

		return result;
	}

	/**
	 * Creates an ontology consisting of the rules that KAON2's conversion to
	 * disjunctive datalog creates. If there is a disjunction in the head,
	 * it is changed into a conjunction. The result may be applied for approximate
	 * reasoning.
	 * 
	 * @param ontology the ontology to be converted
	 * @return the screech program, as an ontology
	 * @throws KAON2Exception from KAON2
	 * @throws InterruptedException from KAON2
	 */
	public Ontology screech(Ontology ontology) throws KAON2Exception, InterruptedException {
		Reasoner reasoner = ontology.createReasoner();
		Collection<Rule> rules = reasoner.getReductionToDisjunctiveDatalog(true, true, true, true);
		Collection<Rule> screechrules = screech(rules);

		DefaultOntologyResolver resolver = new DefaultOntologyResolver();
		resolver.registerReplacement("http://screech/","file:screechtest.xml");
		ontology.getConnection().setOntologyResolver(resolver);
		Ontology result = ontology.getConnection().createOntology("http://screech/", new HashMap<String,Object>());

		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		for (Rule rule : screechrules) {
			// TODO: wouldn't it be easier to switch off ABox-processing in getReductionToDisjunctiveDatalog above?
			if (rule.getBodyLength() > 0) // no rules with emtpy body, instead we need literals (see below)
				changes.add(new OntologyChangeEvent(rule, OntologyChangeEvent.ChangeType.ADD));
		}
		boolean dubexdefault = result.getDuplicateAxiomsThrowException();
		result.setDuplicateAxiomsThrowException(false);
		result.applyChanges(changes);
		changes.clear();

		// add the facts here, since we removed the bodyless facts up there
		Cursor<ClassMember> cmcursor = ontology.createAxiomRequest(ClassMember.class).openCursor();
		while (cmcursor.hasNext()) {
			ClassMember axiom = cmcursor.next();
			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));
		}
		Cursor<ObjectPropertyMember> opmcursor = ontology.createAxiomRequest(ObjectPropertyMember.class).openCursor();
		while (opmcursor.hasNext()) {
			ObjectPropertyMember axiom = opmcursor.next();
			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));
		}
		Cursor<DataPropertyMember> dpmcursor = ontology.createAxiomRequest(DataPropertyMember.class).openCursor();
		while (dpmcursor.hasNext()) {
			DataPropertyMember axiom = dpmcursor.next();
			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));
		}
		result.applyChanges(changes);

		result.setDuplicateAxiomsThrowException(dubexdefault);

		if (m_stat) printer.verbose(", " + api.count(result) + " Rules");
		return result;
	}

	/**
	 * Main processing method. Loads and transforms ontology based on the values
	 * of the various options as given by the class member fields.
	 */
	public void compile() {
		try {
			if (m_format==Format.SWRL) {
				if (!m_output.equals("")) printer.setFile(m_output);
				Ontology ontology = loadOntology(m_fileuri);
				Ontology result = dlpconvert(ontology);
				result.saveOntology(OntologyFileFormat.OWL_RDF, printer.writer, "UTF-8");
				return;
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or reduce ontology " + m_fileuri);
			printer.error(e.toString());
		} catch (InterruptedException i) {
			printer.error("Could not reduce ontology " + m_fileuri);
			printer.error(i.toString());
		} catch (IOException i) {
			printer.error("Could not save to ontology.");
			printer.error(i.toString());
		}

		try {
			ProgramSerializer serializer;

			boolean hasfunctions = false; // can our output handle functions? (simplifies ABox-handling)
			switch (m_format) {
				case Rule     : serializer = new RuleSerializer(printer); break;
				case RuleML   : serializer = new RuleMLSerializer(printer); break;
				case FLogic   :
					serializer = new FLogicSerializer(printer); 
					hasfunctions = true;
				break;
				case LPFLogic :
					serializer = new PrologSerializer(printer, true, true); 
					hasfunctions = true;
				break;
				case Prolog   : 
					serializer = new PrologSerializer(printer, true, false);
					hasfunctions = true;
				break;
				default       : serializer = new PrologSerializer(printer, false, false);
			}

			serializer.setStyle(m_nameStyle);
			serializer.start(m_output);

			try {

				Ontology ontology = loadOntology(m_fileuri);
				Reasoner reasoner = ontology.createReasoner();

				Collection<Rule> rules;
				if (m_allowDDL) {
					//System.out.println(m_allowDDL);
					//System.out.println(m_aBox);
					//System.out.println(m_folding);
					//System.out.println(m_folding && m_aBox);
					rules = reasoner.getReductionToDisjunctiveDatalog(!hasfunctions && m_aBox, m_folding, !hasfunctions && m_folding && m_aBox, true);
				} else {
					rules = reasoner.getReductionToDLP(!hasfunctions && m_aBox, m_folding, !hasfunctions && m_folding && m_aBox, true);
				}
				reasoner.dispose();

				// manually add the ABox and axioms needed for processing
				if (hasfunctions && m_aBox) {
					rules = addABox(rules, ontology);
				}

				// post-transformations:
				rules = addTransitivity(rules, ontology); // add transitivity
				rules = transformICs(rules); // handle ICs
				if (m_screech) { // screech
					rules = screech(rules);
				}
				if (m_single) {
					rules = removeUnfoundedRules(rules);
				}
				rules = sortRules(rules); // sort rules

				// print rules
				serializer.setStat(m_stat);
				serializer.setIC(m_ichandling);
				for (Rule rule : rules) {
					serializer.visit(rule);
				}

				// FIXME: can a disjunctive rule ever be encountered at this point?
				// AND if so, do we want to abort, given that the serializer already went
				// over it? Or should a serializer only record DDL if it is considered a 
				// problem?
				if ( (!m_allowDDL) && (serializer.encounteredDDL()) )
					throw new KAON2Exception("Encountered rule with disjunctive head.");
				if (serializer.hasNegation())
					throw new KAON2Exception("Encountered negative predicate.");

			} catch (KAON2Exception e) {
				printer.error("Could not open or reduce ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not reduce ontology " + m_fileuri);
				printer.error(i.toString());
			}
			finally {
				serializer.end();
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
		
	}

	//////////////////////////////////////////////////////////////////////////////
	//////////
	//////////  Web Interface Methods
	//////////
	//////////////////////////////////////////////////////////////////////////////

	/**
	 * Creates a return string for the Web Interface of DLPConvert.
	 * Only used by the web interface.
	 * 
	 * @param printer used printer
	 * @return stringt to be put into the webpage
	 */
	static String createReturnString(Printer printer) {
		String output = printer.getOutput();
		if (printer.error.equals(""))
			return output;
		String errorheader = "<h2>Error</h2>";
		if (output.equals(""))
			return errorheader + printer.error;
		errorheader += "</pre>The following error occured. dlpconvert tried to continue its work nevertheless and below the errormessage you see the result, but it may be incorrect or incomplete.<br /><pre>";
		return errorheader + printer.error + "<hr />" + output; 
	}
	
	/**
	 * Serializes the result in Flogic.
	 * Only used by the Web Interface. 
	 * 
	 * @param uri input uri of the ontology to be converted
	 * @return the returnstring for the answer HTML-page
	 */
	public static String compileToFlogic(String uri) {
		Printer printer = new Printer();
		printer.setString();
		DLPConvert dlp = new DLPConvert();
		dlp.printer = printer;
		dlp.m_nameStyle = NameSerializationStyle.fragment;
		dlp.m_format = Format.FLogic;
		dlp.m_fileuri = uri;
		dlp.compile();
		return createReturnString(printer);
	}

	/**
	 * Creates a DLPConversion based on an URI.
	 * Only used by the Web Interface.
	 * 
	 * @param uri uri of the ontology to be converted
	 * @param dir where to archive the requeste
	 * @return the return string for the answering HTML-page
	 */
	public static String compileUri(String uri, String dir) {
		Printer printer = new Printer();
		printer.html = true;
		printer.setString();
		count += 1;
		try {
			String filename = dir + "/uri" + String.valueOf(count) + ".uri";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(uri);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		DLPConvert dlp = new DLPConvert();
		dlp.printer = printer;
		dlp.m_nameStyle = NameSerializationStyle.fragment;
		dlp.m_fileuri = uri;
		dlp.compile();
		return createReturnString(printer);
	}
	
	/**
	 * Reads a string as an ontology and then converts this ontology.
	 * Only used by the Web Interface
	 * 
	 * @param string the ontology in string form
	 * @param dir the archive directory
	 * @return the resulting string for the answering HTML-page
	 */
	public static String compileString(String string, String dir) {
        Printer printer = new Printer();
		printer.html = true;
		printer.setString();
        count += 1;
		try {
			String filename = dir + "/owl" + String.valueOf(count) + ".xml";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(string);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		try {
			// construct ontology from string
			KAON2Connection connection = KAON2Manager.newConnection();
			String uri = "http://dlp.aifb.uni-karlsruhe.de/dlpconvert/internal";
			DefaultOntologyResolver resolver=new DefaultOntologyResolver();
			resolver.registerReplacement(uri, "file:temp.owl");
			connection.setOntologyResolver(resolver);
			Ontology ontology = connection.createOntology(uri, new HashMap<String,Object>());

			// import String into ontology
			Reader text = new StringReader(string);
			//OWLImporter importer = new OWLImporter();
			ProgressListener listener = new DefaultOpenProgressListener();
			ontology.importContentsFrom(text, uri, listener);

			// compile it the usual way
			ProgramSerializer serializer = new PrologSerializer(printer);
			serializer.setStyle(ProgramSerializer.NameSerializationStyle.fragment);
			serializer.start("");
			//DLPConvert dlp = new DLPConvert();
			org.semanticweb.kaon2.api.reasoner.Reasoner reasoner = ontology.createReasoner();
			Collection<Rule> rules = reasoner.getReductionToDLP(true, true, true, true);
			for (Rule rule : rules)
				serializer.visit(rule);
			serializer.end();

			// return the compilated
			connection.close();
		} catch (KAON2Exception e) {
			printer.error(e.toString());
		} catch (InterruptedException i) {
			printer.error(i.toString());
		} catch (IOException io) {
			printer.error(io.toString());
		}
		// TODO refactor this whole file
		return createReturnString(printer);
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl dlpconvert <input_url> [<output_filename>] [-atom|-URI|-short] ...");
		// TODO erroresolving, just species
		System.out.println("");
		System.out.println("dlpconvert is a KAON2 based tool that transforms DLP statements from");
		System.out.println("one syntactical representation into another (may be shortened as 'dlp')");
		//System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read. Has to be in OWL/XML format.");
		//System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the transformation will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		//System.out.println("");
		System.out.println("Possible options (take one option from each group):");
		//System.out.println("");
		System.out.println(" -version  prints the version of the tool");
		System.out.println("");
		System.out.println(" -datalog  the output is as a disjunctive datalog program (default)");
		System.out.println(" -prolog   the output is as a valid Prolog program, disjunctions are dropped.");
		System.out.println(" -flogic   the output is F-Logic, encoded using F-Molecules and class hierarchies.");
		System.out.println(" -lpflogic the output is F-Logic, encoded using quantified logic programming rules.");
		System.out.println(" -swrl     the output is OWL+SWRL");
		System.out.println(" -ruleml   the output is RuleML");
		System.out.println(" -kaon2    the output is in the internal KAON2 format");
		System.out.println("");
		System.out.println(" -URI      the IDs are full qualified URIs (default)");
		System.out.println(" -atom     the names are transformed to legal XSB atoms (not with -kaon2)");
		System.out.println(" -short    just the fragment identifiers  (most readable, not with -kaon2)");
		// TODO System.out.println(" -ns       namespace and fragment identifier combined");
		System.out.println("");
		System.out.println(" -single   optimise program for stand-alone use; remove all parts that are not");
		System.out.println("           needed if no further data or rules are added.");
		System.out.println(" -aux      add all necessary definitions of KAON2 auxilliary predicates (default)");
		System.out.println(" -noaux    do not add definitions of KAON2 auxilliary predicates.");
		System.out.println("");
//		System.out.println(" -ABox     processes the ABox (default)");
//		System.out.println(" -noABox   skips the ABox Processing");
//		System.out.println("");
//		System.out.println(" -unfold   unfolds the new function terms introduced in reduction (default)");
//		System.out.println(" -nounfold skips the unfolding");
//		System.out.println("");
		System.out.println(" -X        allows disjunctions in head (uses disjunctive datalog algorihtm)");
		System.out.println(" -DLP      disallows disjunctions in head (default)");
		System.out.println(" -screech  breaks disjunctions in head to multiple rules (sets DDL)");
		System.out.println("");
		System.out.println(" -IC       leaves integrity constraints in resulting program (default)");
		System.out.println(" -noIC     removes integrity constraints from result");
		System.out.println(" -ICP      (or -ICPredicate) use predicate 'Inconsistent' instead of empty head");
		System.out.println("           (default for -prolog and -flogic)");
		System.out.println("");
		System.out.println(" -verbose  offers further information (off by default)");
	}
}
