/*
 * deo weakens ontologies by transforming nominals in simple classes.
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;

/**
 * Deo weakens ontologies by transforming nominals in simple classes.
 */
public class Deo extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "deo"; }
	
	/**
	 * Random variable. Random numbers are taken for the not so clever naming algorithm
	 * for the new simple classes.
	 */
	private Random m_random = new Random();
	
	/**
	 * The URI of the ontology to be weakened
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to save the ontology
	 */
	private String m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, the result will be supressed 
	 */	
	private boolean m_result = true;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Deo d = new Deo();
		if (!d.processArguments(args)) return;
		d.compile();
		d.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("deo"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];
			
			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.FUNCTIONAL_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("deo")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the ded owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {

	        	Ontology ontology = loadOntology(m_fileuri);
	        	ontology = deo(ontology);
	        	
	        	if (m_result)
	        		ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	
	/**
	 * Creates a new URI
	 * @return a new URI
	 */
	String newURI() {
		// TODO make a more elaborate algorithm
		return "http://www.example.com/formerNominal#" + Math.abs(m_random.nextInt()); 
	}

	/**
	 * Takes a description and changes it in such a way that it does not hold any nominals,
	 * but instead simple individuals and an appropriate class with the membership of all
	 * this individuals.
	 * @param description Description to be changed
	 * @param ontology Ontology being weakened
	 * @param changes Changed that have to happen
	 * @param owlClass If there is already a class to which the individuals have to belong, it is given here
	 * @return Changed description
	 * @throws KAON2Exception
	 */
	Description processDescription(Description description, Ontology ontology, List<OntologyChangeEvent> changes, OWLClass owlClass) throws KAON2Exception {
		Description result = null;
		
		if (ObjectOneOf.class.isInstance(description)) {
			ObjectOneOf nominal = (ObjectOneOf)description;
			if (owlClass == null)
				owlClass = KAON2Manager.factory().owlClass(newURI());
			Set<Individual> individuals = nominal.getIndividuals();

	        for (Individual individual : individuals) {
				ClassMember classMember = KAON2Manager.factory().classMember(owlClass, individual);
	        	if (!ontology.containsAxiom(classMember, false)) {
					printer.verbose("Adding: " + classMember.toString());
			        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
	        	} else {
	        		printer.verbose("Already in: " + classMember.toString());
	        	}
			}
			result = owlClass;
		}
		
		if (ObjectHasValue.class.isInstance(description)) {
			ObjectHasValue hasValue = (ObjectHasValue)description;

			owlClass = KAON2Manager.factory().owlClass(newURI());
			Individual individual = hasValue.getIndividual();
			
			ClassMember classMember = KAON2Manager.factory().classMember(owlClass, individual);
        	if (!ontology.containsAxiom(classMember, false)) {
				printer.verbose("Adding: " + classMember.toString());
		        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
        	} else {
        		printer.verbose("Already in: " + classMember.toString());
	        }

	        ObjectAll objectAll = KAON2Manager.factory().objectAll(hasValue.getObjectProperty(), owlClass);
	        result = objectAll;
		}
		
		if (ObjectAll.class.isInstance(description)) {
			ObjectAll all = (ObjectAll)description;
			
			Description allFrom = processDescription(all.getDescription(), ontology, changes, null);
			if (allFrom!=null) {
				ObjectAll objectAll = KAON2Manager.factory().objectAll(all.getObjectProperty(), allFrom);
				result = objectAll;
			}
		}
		
		if (ObjectSome.class.isInstance(description)) {
			ObjectSome some = (ObjectSome)description;
			
			Description someFrom = processDescription(some.getDescription(), ontology, changes, null);
			if (someFrom!=null) {
				ObjectSome objectSome = KAON2Manager.factory().objectSome(some.getObjectProperty(), someFrom);
				result = objectSome;
			}
		}
		
		if (ObjectCardinality.class.isInstance(description)) {
			ObjectCardinality cardinality = (ObjectCardinality)description;
			
			Description cardinalityOn = processDescription(cardinality.getDescription(), ontology, changes, null);
			if (cardinalityOn!=null) {
				ObjectCardinality objectCardinality = KAON2Manager.factory().objectCardinality(cardinality.getCardinalityType(), cardinality.getCardinality(), cardinality.getObjectProperty(), cardinalityOn);
				result = objectCardinality;
			}
		}
		
		if (ObjectAnd.class.isInstance(description)) {
			ObjectAnd and = (ObjectAnd)description;
			
			Set<Description> descriptions = and.getDescriptions();
			Set<Description> newDescriptions = new HashSet<Description>();
			boolean unchanged = true;
			for (Description element : descriptions) {
				Description processed = processDescription(element, ontology, changes, null);
				if (processed == null) {
					newDescriptions.add(element);
				} else {
					newDescriptions.add(processed);
					unchanged = false;
				}
			}
			ObjectAnd objectAnd = KAON2Manager.factory().objectAnd(newDescriptions);
			if (!unchanged)
				result = objectAnd;
		}
		
		if (ObjectOr.class.isInstance(description)) {
			ObjectOr or = (ObjectOr)description;
			
			Set<Description> descriptions = or.getDescriptions();
			Set<Description> newDescriptions = new HashSet<Description>();
			boolean unchanged = true;
			for (Description element : descriptions) {
				Description processed = processDescription(element, ontology, changes, null);
				if (processed == null) {
					newDescriptions.add(element);
				} else {
					newDescriptions.add(processed);
					unchanged = false;
				}
			}
			ObjectOr objectOr = KAON2Manager.factory().objectOr(newDescriptions);
			if (!unchanged)
				result = objectOr;
		}
		
		if (ObjectNot.class.isInstance(description)) {
			ObjectNot not = (ObjectNot)description;
			
			Description isNot = processDescription(not.getDescription(), ontology, changes, null);
			if (isNot!=null) {
				ObjectNot objectNot = KAON2Manager.factory().objectNot(isNot);
				result = objectNot;
			}
		}
		
		
		return result;
	}
	
	/**
	 * Takes an ontology and replaces all the nominals inside by simple class membership
	 * @param ontology Ontology to be weakened
	 * @return weakened ontology
	 * @throws KAON2Exception
	 */
	public Ontology deo(Ontology ontology) throws KAON2Exception {
		// look for everything that has a description
		
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
  		
  		Cursor<ClassMember> classCursor = ontology.createAxiomRequest(ClassMember.class).openCursor();
  		while (classCursor.hasNext()) {
  			ClassMember a = classCursor.next();
  			Description c = a.getDescription();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ClassMember classMember = KAON2Manager.factory().classMember(result, a.getIndividual());
		        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + classMember.toString());
		        printer.verbose("Deleting: " + a.toString());
	  		}
  		}
  		classCursor.close();

  		Cursor<DataPropertyDomain> dataPropertyDomainCursor = ontology.createAxiomRequest(DataPropertyDomain.class).openCursor();
  		while (dataPropertyDomainCursor.hasNext()) {
  			DataPropertyDomain a = dataPropertyDomainCursor.next();
  			Description c = a.getDomain();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				DataPropertyDomain dataPropertyDomain = KAON2Manager.factory().dataPropertyDomain(a.getDataProperty(), result);
		        changes.add(new OntologyChangeEvent(dataPropertyDomain, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + dataPropertyDomain.toString());
		        printer.verbose("Deleting: " + a.toString());
	  		}
  		}
  		dataPropertyDomainCursor.close();

  		Cursor<ObjectPropertyDomain> objectPropertyDomainCursor = ontology.createAxiomRequest(ObjectPropertyDomain.class).openCursor();
  		while (objectPropertyDomainCursor.hasNext()) {
  			ObjectPropertyDomain a = objectPropertyDomainCursor.next();
  			Description c = a.getDomain();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ObjectPropertyDomain objectPropertyDomain = KAON2Manager.factory().objectPropertyDomain(a.getObjectProperty(), result);
		        changes.add(new OntologyChangeEvent(objectPropertyDomain, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + objectPropertyDomain.toString());
		        printer.verbose("Deleting: " + a.toString());
	  		}
  		}
  		objectPropertyDomainCursor.close();

  		Cursor<ObjectPropertyRange> objectPropertyRangeCursor = ontology.createAxiomRequest(ObjectPropertyRange.class).openCursor();
  		while (objectPropertyRangeCursor.hasNext()) {
  			ObjectPropertyRange a = objectPropertyRangeCursor.next();
  			Description c = a.getRange();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ObjectPropertyRange objectPropertyRange = KAON2Manager.factory().objectPropertyRange(a.getObjectProperty(), result);
		        changes.add(new OntologyChangeEvent(objectPropertyRange, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + objectPropertyRange.toString());
		        printer.verbose("Deleting: " + a.toString());
	  		}
  		}
  		objectPropertyRangeCursor.close();

  		Cursor<SubClassOf> subClassOfCursor = ontology.createAxiomRequest(SubClassOf.class).openCursor();
  		while (subClassOfCursor.hasNext()) {
  			SubClassOf a = subClassOfCursor.next();
  			Description c = a.getSubDescription();
	  		Description sub = processDescription(c, ontology, changes, null);
  			c = a.getSuperDescription();
	  		Description sup = processDescription(c, ontology, changes, null);
	  		if ((sub!=null)||(sup!=null)) {
	  			if (sub==null) sub = a.getSubDescription();
	  			if (sup==null) sup = a.getSuperDescription();
				SubClassOf subClassOf = KAON2Manager.factory().subClassOf(sub, sup);
		        changes.add(new OntologyChangeEvent(subClassOf, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + subClassOf.toString());
		        printer.verbose("Deleting: " + a.toString());
	  		}
  		}
  		subClassOfCursor.close();

  		Cursor<DisjointClasses> disjointClassesCursor = ontology.createAxiomRequest(DisjointClasses.class).openCursor();
  		while (disjointClassesCursor.hasNext()) {
  			DisjointClasses a = disjointClassesCursor.next();
  			Set<Description> cs = a.getDescriptions();
  			Set<Description> newCs = new HashSet<Description>();
  			boolean hasNominal = false;
  			for (Description c : cs) {
		  		Description result = processDescription(c, ontology, changes, null);
		  		if (result==null) {
		  			newCs.add(c);
		  		} else {
		  			newCs.add(result);
		  			hasNominal = true;
		  		}
  			}
  			if (hasNominal) {
				DisjointClasses disjointClasses = KAON2Manager.factory().disjointClasses(newCs);
		        changes.add(new OntologyChangeEvent(disjointClasses, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Adding: " + disjointClasses.toString());
		        printer.verbose("Deleting: " + a.toString());
  			}
  		}
  		disjointClassesCursor.close();

  		Cursor<EquivalentClasses> equivalentClassesCursor = ontology.createAxiomRequest(EquivalentClasses.class).openCursor();
  		while (equivalentClassesCursor.hasNext()) {
  			EquivalentClasses a = equivalentClassesCursor.next();
  			Set<Description> cs = a.getDescriptions();
  			Set<Description> newCs = new HashSet<Description>();
  			boolean hasNominal = false;
  			OWLClass simpleClass = null;

  			//for (Description c : cs) {
  			//	if (OWLClass.class.isInstance(c))
  			//		simpleClass = (OWLClass)c;
  			//}
  			
  			for (Description c : cs) {
		  		Description result = processDescription(c, ontology, changes, simpleClass);
		  		if (result==null) {
		  			newCs.add(c);
		  		} else {
		  			if (simpleClass == null)
		  				newCs.add(result);
		  			hasNominal = true;
		  		}
  			}
  			if (hasNominal) {
  				if (newCs.size() > 1) {
					EquivalentClasses equivalentClasses = KAON2Manager.factory().equivalentClasses(newCs);
			        changes.add(new OntologyChangeEvent(equivalentClasses, OntologyChangeEvent.ChangeType.ADD));
			        printer.verbose("Adding: " + equivalentClasses.toString());
  				}
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        printer.verbose("Deleting: " + a.toString());
  			}
  		}
  		equivalentClassesCursor.close();
    	
        ontology.applyChanges(changes);
 	
	  	return ontology;
	}
	
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl deo <input_url> [<output_filename>] [-verb] [-rdf|-xml] [-noresult]");
		System.out.println("");
		System.out.println("deo weakens an ontology by transforming all nominals into simple classes.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the weakened ontology will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result. In combination with -verbose you");
		System.out.println("     get a list of the needed changes, without apllying them.");
	}
	
}
