/*
 * Created on 24.03.2005
 *
 * Diff prints all axioms which are contained in the second, but
 * not in the first ontology.
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;

/**
 * Returns all axioms that are within one, but not within the other ontology
 */
public class Diff extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "diff"; }

	/**
	 * The ontology URI of the base file
	 */
	private String m_fileuriBase = "";
	
	/**
	 * The ontology URI of the file that is to be substracted
	 */
	private String m_fileuriDiff = "";
	
	/**
	 * The name of the output file
	 */
	private String m_output = "";
	
	/**
	 * The file format the result is to be serialized in
	 */
	private String m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * if false, no output will be created
	 */
	private boolean m_result = true;

	/**
	 * If the tool is being called from the outside, some basic processing
	 * is being done
	 * @param args Command line parameters. The same as for the OWLtools call
	 */
	public static void main(String[] args) {
		Diff d = new Diff();
		if (!d.processArguments(args)) return;
		d.compile();
		d.printer.close();
	}

	/**
	 * processes the command line parameteres
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
		if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("diff"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName1 = false;
		boolean hasFileName2 = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.FUNCTIONAL_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("diff")) {
					unknownArg = false;
				} else {
					if (hasFileName1) {
						if (hasFileName2) {
							if (hasOutput) { printUsageError(); return false; }
							m_output = arg;
							hasOutput = true;
							unknownArg = false;
						} else {
							m_fileuriDiff = arg;
							hasFileName2 = true;					
							unknownArg = false;
						}
					} else {
						m_fileuriBase = arg;
						hasFileName1 = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; }
		}
		
		if (!hasFileName2) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the Dump owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology base = loadOntology(m_fileuriBase);
	        	Ontology diff = loadOntology(m_fileuriDiff);
	        	Ontology result = diff(base, diff);
	        	
	        	if (m_result)
	        		result.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology.");
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology.");
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	
	/**
	 * Creates an ontology for the output instead needing to be supplied with one.
	 * That ontology will hold all the axioms that are in the diff ontology, but not
	 * in the base ontology
	 * @param base The ontology to compare to
	 * @param diff The ontology that has the extra axioms
	 * @return The difference between the two input ontologies
	 * @throws KAON2Exception
	 */
	public Ontology diff(Ontology base, Ontology diff) throws KAON2Exception {
		Ontology result = null;
		result = createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/diff/result");
		diff(base, diff, result);
		return result;
	}
	
	/**
	 * Saves all the axioms that are in the diff ontology but not in base ontology in
	 * the result ontology
	 * @param base The ontology to compare to
	 * @param diff The ontology with the extra axioms
	 * @param result The extra axioms are added to the result ontology
	 * @throws KAON2Exception
	 */
	public void diff(Ontology base, Ontology diff, Ontology result) throws KAON2Exception {
  		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
  		
  		Cursor<Axiom> cursor = diff.createAxiomRequest().openCursor();
  		while (cursor.hasNext()) {
  			Axiom axiom = cursor.next();
  			if (!base.containsAxiom(axiom, false)) {
		        changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));	  				
  				printer.verbose(axiom.toString());
  			}
  		}

  		result.applyChanges(changes);
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl diff <url1> <url2> [<output_filename>] [-verbose] [-rdf|-xml] [-no]");
		System.out.println("");
		System.out.println("diff returns all axioms given in the second ontology not existing in the first.");
		System.out.println("");
		System.out.println(" <url1>");
		System.out.println("     The location of the base ontology");
		System.out.println(" <url2>");
		System.out.println("     The location of the compared ontology");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where ontology with the differing axioms will be");
		System.out.println("     written to. If omitted, the result is sent to the console.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result.");
	}	
}
