/*
 * filter removes axioms of a specific kind
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;

/**
 * Filter removes axioms from an ontology, specified by their axiom type
 */
public class Filter extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "filter"; }
	
	/**
	 * The URI of the ontology to be filtered
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to save the ontology
	 */
	private String m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, the result will be supressed 
	 */
	private boolean m_result = true;
	
	/**
	 * The class to be filtered by
	 */
	private Class m_filter = null;
	
	/**
	 * If to remove the specified type -- or the rest
	 */
	private boolean m_only = false;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Filter filter = new Filter();
		if (!filter.processArguments(args)) return;
		filter.compile();
		filter.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("filter"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		boolean hasFilter = false;
		boolean hasOnly = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];
			
			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				if (arg.equalsIgnoreCase("-only")) {
					if (hasOnly) { printUsageError(); return false; }
					m_only = true;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.FUNCTIONAL_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}

				// filter
				if (arg.equalsIgnoreCase("-Annotation")||arg.equalsIgnoreCase("-a")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = EntityAnnotation.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ClassMember")||arg.equalsIgnoreCase("-cm")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ClassMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyAttribute")||arg.equalsIgnoreCase("-dpa")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DataPropertyAttribute.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyDomain")||arg.equalsIgnoreCase("-dpd")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DataPropertyDomain.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyMember")||arg.equalsIgnoreCase("-dpm")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DataPropertyMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyRange")||arg.equalsIgnoreCase("-dpr")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DataPropertyRange.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DifferentIndividuals")||arg.equalsIgnoreCase("-di")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DifferentIndividuals.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DisjointClasses")||arg.equalsIgnoreCase("-dc")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DisjointClasses.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentClasses")||arg.equalsIgnoreCase("-ec")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = EquivalentClasses.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentDataProperties")||arg.equalsIgnoreCase("-edp")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = EquivalentDataProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentObjectProperties")||arg.equalsIgnoreCase("-eop")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = EquivalentObjectProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-InverseObjectProperties")||arg.equalsIgnoreCase("-iop")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = InverseObjectProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyAttribute")||arg.equalsIgnoreCase("-opa")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectPropertyAttribute.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyDomain")||arg.equalsIgnoreCase("-opd")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectPropertyDomain.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyMember")||arg.equalsIgnoreCase("-opm")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectPropertyMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyRange")||arg.equalsIgnoreCase("-opr")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectPropertyRange.class;
					hasFilter = true;
					unknownArg = false;
				}
//				if (arg.equalsIgnoreCase("-Rule")||arg.equalsIgnoreCase("-r")) {
//					if (hasFilter) { printUsageError(); return false; }
//					m_filter = Rule.class;
//					hasFilter = true;
//					unknownArg = false;
//				}
				if (arg.equalsIgnoreCase("-SameIndividual")||arg.equalsIgnoreCase("-si")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = SameIndividual.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubClassOf")||arg.equalsIgnoreCase("-sco")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = SubClassOf.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubDataPropertyOf")||arg.equalsIgnoreCase("-sdpo")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = SubDataPropertyOf.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubObjectPropertyOf")||arg.equalsIgnoreCase("-sopo")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = SubObjectPropertyOf.class;
					hasFilter = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("filter")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFilter) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
		  		if (m_only)
		  			ontology = filterOut(ontology, m_filter);
		  		else
		  			ontology = filter(ontology, m_filter);
	        	if (m_result)
	        		ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	
	/**
	 * Removes all axioms of a certain kind from an ontology
	 * @param ontology The input ontology
	 * @param filterclass The axiom type to be removed
	 * @return The ontology without the removed axioms
	 * @throws KAON2Exception
	 */
	public Ontology filterOut(Ontology ontology, Class filterclass) throws KAON2Exception {
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
  		Cursor<Axiom> cursor = ontology.createAxiomRequest(filterclass).openCursor();
  		while (cursor.hasNext()) {
  			Axiom axiom = cursor.next();
  			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.REMOVE));
  			printer.verbose("Removing " + axiom.toString());
  		}
  		cursor.close();
   		ontology.applyChanges(changes);
        return ontology;
	}

	/**
	 * Removes all axioms but those of the specified kind from an ontology
	 * @param ontology The ontology to be filtered
	 * @param filterclass The type of axioms to remain in the ontology
	 * @return The filtered ontology
	 * @throws KAON2Exception
	 */
	public Ontology filter(Ontology ontology, Class filterclass) throws KAON2Exception {
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
  		Cursor<Axiom> cursor = ontology.createAxiomRequest().openCursor();
  		while (cursor.hasNext()) {
  			Axiom axiom = cursor.next();
  			if (!filterclass.isInstance(axiom)) {
	  			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.REMOVE));
	  			printer.verbose("Removing " + axiom.toString());
  			}
  		}
  		cursor.close();
        ontology.applyChanges(changes);
        return ontology;
	}
		
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl filter <input_url> [<output_filename>] [-{filter}] [options]");
		System.out.println("");
		System.out.println("filter removes all Axioms of a specific kind from an ontology (or fetches them)");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the filtered ontology will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println(" -only");     
		System.out.println("     Instead of deleting all filtered axioms, it keeps only those filtered.");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result. In combination with -verbose you");
		System.out.println("     get a list of the needed changes, without apllying them.");
		System.out.println(" -{filter}");
		System.out.println("     sets the kind of axioms to be filtered. Available filters: Annotation,");
		System.out.println("     ClassMember, DataPropertyAttribute, DataPropertyDomain,DataPropertyMember,");
		System.out.println("     DataPropertyRange, DifferentIndividuals, DisjointClasses,");
		System.out.println("     EquivalentClasses, EquivalentDataProperties, EquivalentObjectProperties,");
		System.out.println("     InverseObjectProperties, ObjectPropertyAttribute, ObjectPropertyDomain,");
		System.out.println("     ObjectPropertyMember,ObjectPropertyRange, Rule,SameIndividual, SubClassOf,");
		System.out.println("     SubDataPropertyOf, SubObjectPropertyOf. All kinds may be abbreviated with");
		System.out.println("     the capital letters (e.g. CM instead ClassMember)");
	}
}
