/*
 * Created on 23.03.2005
 *
 * Offers some basic functionality common to the tools
 */
package edu.unika.aifb.owltools.tools;

import java.util.*;
import org.semanticweb.kaon2.api.*;

import edu.unika.aifb.owltools.utils.Printer;

/**
 * Common code for all the owl tools
 */
abstract public class KAON2Tool {
	
	/**
	 * The KAON2Connection for the tool
	 */
	protected KAON2Connection m_connection = null;
	
	/**
	 * The printer for the different kinds of output. This is public so when used
	 * in own code the printer can be redirected in any way
	 */
	public Printer printer = null;
	
	/**
	 * Constructor. Basically creates a new printer. Most of the actual constructing
	 * happens when processing the command line arguments later.
	 *
	 */
	public KAON2Tool() {
		printer = new Printer();
	}
	
	/**
	 * the name of the specified tool. Needs to be implemented by all tools
	 * 
	 * @return the name of the tool
	 */
	abstract public String name();
	
	/**
	 * needs to be implemented by all tools:
	 * processArgument to process all the given command line arguments
	 */
	abstract public boolean processArguments(String[] args);

	/**
	 * Creates a new connection if there is none yet 
	 * @throws KAON2Exception
	 */
	void makeConnection() throws KAON2Exception {
		if (m_connection == null)
			m_connection = KAON2Manager.newConnection();
	}
	
	/**
	 * closes an existing KAON2Connection
	 * @throws KAON2Exception
	 */
	protected void endConnection() throws KAON2Exception {
		if (m_connection != null)
			m_connection.close(); // GANZ HEFTIGES TODO memory leak?
		m_connection = null;
	}
	
	/**
	 * loads an ontology. If there is no open connection, a new one is created
	 * @param name Name of the ontology to be loaded. Either starts with http, https or ftp, or it will be loaded from a local file with the given name
	 * @return The loaded ontology
	 * @throws KAON2Exception
	 * @throws InterruptedException
	 */
	protected Ontology loadOntology(String name) throws KAON2Exception, InterruptedException {
		makeConnection();
		
		OntologyResolver resolver = m_connection.getOntologyResolver();
		if (DefaultOntologyResolver.class.isInstance(resolver)) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
	        if (!name.startsWith("http:"))
		        if (!name.startsWith("https:"))
			        if (!name.startsWith("ftp:"))
			        	if (!name.startsWith("file:"))
			        		res.registerReplacement(name, "file:" + name);
	        resolver = res;
		}
        m_connection.setOntologyResolver(resolver);
		
       	Ontology ontology = m_connection.openOntology(name, new HashMap<String,Object>());
       	return ontology;
	}
	
	/**
	 * Creates a new ontology
	 * @param name the URI for the new ontology
	 * @return the newly created ontology
	 * @throws KAON2Exception
	 */
	Ontology createOntology(String name) throws KAON2Exception {
		OntologyResolver resolver = m_connection.getOntologyResolver();
		if (DefaultOntologyResolver.class.isInstance(resolver)) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
			res.registerReplacement(name, "file:temp.owl");
	        resolver = res;
		}
        m_connection.setOntologyResolver(resolver);

        Ontology ontology = m_connection.createOntology(name, new HashMap<String, Object>());
		return ontology;
	}
	
	/**
	 * prints a short excuse after not understandable input
	 */
	public void printUsageError() {
		System.out.println("Sorry, didn't understand the command line parameters. Please try entering:");
		System.out.println("owl " + name() + " help");
	}	
	
}
