package edu.unika.aifb.owltools.tools;

import java.io.IOException;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.formatting.*;

/**
 * Asserts that all individuals in an ontology are different from each other 
 */
public class Luna extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "luna"; }
	
	/**
	 * The URI of the ontology to be filtered
	 */
	private String m_fileuri = "";
	
	/**
	 * The output filename for the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to save the ontology
	 */
	private String m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, the result will be supressed 
	 */
	private boolean m_result = true;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Luna luna = new Luna();
		if (!luna.processArguments(args)) return;
		luna.compile();
		luna.printer.close();
	}

	/**
	 * Processes the command line arguments
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("luna"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		boolean hasVerbose = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {				
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					

				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.FUNCTIONAL_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
				
			} else {
				if (arg.equalsIgnoreCase("luna")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}
	
	/**
	 * Adds an axiom to the ontology that states that all individuals
	 * are different from each other.
	 * @param ontology The ontology that needs to get the new axiom
	 * @return The changed ontology
	 * @throws KAON2Exception
	 */
	public Ontology luna(Ontology ontology) throws KAON2Exception {
		Request<Individual> request = ontology.createEntityRequest(Individual.class);
		Set<Individual> individuals = request.get();
		Axiom disAxiom = KAON2Manager.factory().differentIndividuals(individuals);
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
        changes.add(new OntologyChangeEvent(disAxiom, OntologyChangeEvent.ChangeType.ADD));
        printer.verbose("Adding axiom: " + disAxiom.toString());
        ontology.applyChanges(changes);
		return ontology;
	}

	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
	  			ontology = luna(ontology);

        		if (m_result)
        			ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl luna <input_url> [<output_filename>] [options]");
		System.out.println("");
		System.out.println("luna (local Unique Name Assumption) adds an axiom to an ontology, stating the");
		System.out.println("difference of all named instances individuals within the ontology.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the resulting ontology will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
	}
}
