/*
 * Satisiable checks satisfiability with KAON2
 */
package edu.unika.aifb.owltools.tools;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.reasoner.*;

/**
 * Satisiable loads an ontology and lets KAON2 check if it is satisfiable
 */
public class Satisfiable extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "satisiable"; }

	/**
	 * The URI of the ontology to be filtered
	 */
	private String m_fileuri = "";

	/**
	 * The output filename for the result (in case anyone prefers a
	 * file with the result instead of a simple command line yes)
	 */
	private String m_output = "";
	
	/**
	 * If the time should be taken or not
	 */
	private Boolean m_time = false;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Satisfiable s = new Satisfiable();
		if (!s.processArguments(args)) return;
		s.compile();
		s.printer.close();
	}

	/**
	 * process command line arguments
	 */
	public boolean processArguments(String[] args) {
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("satisfiable")) || (s.equalsIgnoreCase("sat"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				if (arg.equalsIgnoreCase("-time") || arg.equalsIgnoreCase("-t")) {
					m_time = true;
					unknownArg = false;
				}				
			} else {
				if (arg.equalsIgnoreCase("satisfiable") || arg.equalsIgnoreCase("sat")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}
	
	/**
	 * Checks if the fiven ontology is satisfiable
	 * @param ontology The ontology to be checked
	 * @return true if the ontology is satisfiable, else false
	 * @throws KAON2Exception
	 * @throws InterruptedException
	 */
	public boolean satisfiable(Ontology ontology) throws KAON2Exception, InterruptedException {
		Reasoner reasoner = ontology.createReasoner();
		return reasoner.isSatisfiable();
	}
		
	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);
	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
	        	long start, end = 0;
				start = System.currentTimeMillis();
				boolean sat = satisfiable(ontology);
				end = System.currentTimeMillis();	        	
	        	if (sat)
	        		printer.putline("yes");
	        	else
	        		printer.putline("no");
	        	if (m_time)
	        		printer.putline((end-start) + " ms");
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not reason on ontology " + m_fileuri);
				printer.error(i.toString());
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl satisfiable <input_url> [-time]");
		System.out.println("");
		System.out.println("Checks if a given ontology is satisfiable. If so, it returns 'yes', else 'no' ");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println(" -time");
		System.out.println("     Takes the time of how long the satsifiability check took");
	}
}
